### import ####################################################################


import os
import traceback
import imp
import time
import datetime
import dateutil
import collections
import pathlib

from PySide2 import QtCore, QtWidgets

import appdirs
import toml

import WrightTools as wt

import yaqc_cmds.__main__
import yaqc_cmds.project.project_globals as g
import yaqc_cmds.project.classes as pc
import yaqc_cmds.project.widgets as pw
import yaqc_cmds.project.file_dialog_handler as file_dialog_handler

import yaqc_cmds.hardware.spectrometers as spectrometers
import yaqc_cmds.hardware.delays as delays
import yaqc_cmds.hardware.opas as opas
import yaqc_cmds.hardware.filters as filters
import yaqc_cmds.somatic as somatic

all_hardwares = opas.hardwares + spectrometers.hardwares + delays.hardwares + filters.hardwares

### define ####################################################################


app = g.app.read()

main_window = g.main_window.read()

somatic_folder = os.path.dirname(__file__)
data_folder = main_window.data_folder

### queue item classes ########################################################


class Item(QtCore.QObject):
    updated = QtCore.Signal()

    def __init__(self, name="", info="", description=""):
        QtCore.QObject.__init__(self)
        # basic information
        self.name = name  # filled by user, up to 10 charachters long
        self.info = info  # filled by user, can be arbitrarily long
        self.description = description  # filled programmatically, to be displayed in table
        # status: can be one of 'ENQUEUED', 'RUNNING', 'COMPLETE', 'FAILED'
        self.status = "ENQUEUED"
        self.finished = pc.Bool(initial_value=False)
        # timestamps
        self.created = wt.kit.TimeStamp()
        self.started = None
        self.exited = None

    def write_to_ini(self, ini, section):
        pass


class Acquisition(Item):
    def __init__(self, aqn_path, module, **kwargs):
        Item.__init__(self, **kwargs)
        self.aqn_path = aqn_path
        self.type = "acquisition"
        self.module = module
        self.url = None

    def write_to_ini(self, ini, section):
        ini.write(section, "path", self.aqn_path)
        ini.write(section, "url", self.url)


class Device(Item):
    def __init__(self, **kwargs):
        Item.__init__(self, **kwargs)
        self.type = "acquisition"


class Hardware(Item):
    def __init__(self, hardwares, value, units, **kwargs):
        Item.__init__(self, **kwargs)
        self.type = "hardware"
        self.hardwares = [ah for ah in all_hardwares if ah.name in hardwares]
        self.value = value
        self.units = units

    def execute(self):
        for hw in self.hardwares:
            hw.set_position(self.value, self.units)
        g.hardware_waits.wait()

    def write_to_ini(self, ini, section):
        ini.write(section, "value", self.value)
        ini.write(section, "units", self.units)
        ini.write(section, "hardwares", [hw.name for hw in self.hardwares])


class Interrupt(Item):
    def __init__(self, **kwargs):
        Item.__init__(self, **kwargs)
        self.type = "interrupt"


class Script(Item):
    def __init__(self, **kwargs):
        Item.__init__(self, **kwargs)
        self.type = "script"


class Wait(Item):
    def __init__(self, operation, amount, **kwargs):
        Item.__init__(self, **kwargs)
        self.type = "wait"
        self.operation = operation
        self.amount = amount

    def write_to_ini(self, ini, section):
        ini.write(section, "operation", self.operation)
        ini.write(section, "amount", self.amount)


### worker ####################################################################


class Worker(QtCore.QObject):
    action_complete = QtCore.Signal()

    def __init__(self, enqueued, busy, queue_status, queue_url, queue_index, queue_folder):
        QtCore.QObject.__init__(self)
        self.enqueued = enqueued
        self.busy = busy
        self.queue_status = queue_status
        self.fraction_complete = pc.Number(initial_value=0.0)
        self.queue_url = queue_url
        self.index = queue_index
        self.folder = queue_folder

    def check_busy(self, _=[]):
        if self.enqueued.read():  # there are items in enqueued
            time.sleep(0.1)  # don't loop like crazy
            self.busy.write(True)
        else:
            self.busy.write(False)

    @QtCore.Slot(str, list)
    def dequeue(self, method, inputs):
        """
        Slot to accept enqueued commands from main thread.

        Method passed as qstring, inputs as list of [args, kwargs].
        """
        args, kwargs = inputs
        if g.debug.read():
            print("worker dequeue:", method, inputs)
        # the queue should only be adding items to execute
        item = args[0]
        g.queue_control.write(True)
        somatic.signals.queue_taking_control.emit()
        self.queue_status.going.write(True)
        self.fraction_complete.write(0.0)
        item.started = wt.kit.TimeStamp()
        if item.type == "acquisition":
            self.execute_acquisition(item)
        elif item.type == "device":
            self.execute_device(item)
        elif item.type == "hardware":
            self.execute_hardware(item)
        elif item.type == "wait":
            self.execute_wait(item)
        elif item.type == "interrupt":
            self.execute_interrupt(item)
        item.exited = wt.kit.TimeStamp()
        self.fraction_complete.write(1.0)
        self.queue_status.going.write(False)
        g.queue_control.write(False)
        somatic.signals.queue_relinquishing_control.emit()
        self.action_complete.emit()
        # remove item from enqueued
        self.enqueued.pop()
        if not self.enqueued.read():
            self.check_busy([])

    def execute_acquisition(self, item):
        # create acquisition worker object
        module = item.module
        worker = module.Worker(item.aqn_path, self, item.finished)
        item.aqn_path = worker.aqn_path
        # create acquisition folder on google drive
        folder_name = os.path.abspath(item.aqn_path)[:-4]
        if g.google_drive_enabled.read():
            g.google_drive_control.read().reserve_id(folder_name)
            url = g.google_drive_control.read().id_to_open_url(folder_name)
            g.google_drive_control.read().create_folder(
                folder_name, self.folder.read(), folder_name
            )
            ini = wt.kit.INI(item.aqn_path)
            ini.write("info", "url", url)
            item.url = url
        # run it
        try:
            worker.run()
        except Exception as error:
            # TODO: log error
            print("ACQUISITION ERROR:", error)
            traceback.print_exc()
        # upload aqn file
        if g.google_drive_enabled.read():
            g.google_drive_control.read().create_file(
                path=str(item.aqn_path),
                parent_id=self.folder.read(),
                id=str(item.aqn_path),
            )

        # send message on slack
        if g.slack_enabled.read():
            name = os.path.split(folder_name)[1]
            message = ":checkered_flag: acquisition complete - {0} - {1}".format(name, item.url)
            g.slack_control.read().send_message(message)

    def execute_device(self, item):
        # TODO:
        time.sleep(5)
        item.finished.write(False)

    def execute_hardware(self, item):
        item.execute()
        item.finished.write(True)

    def execute_script(self, item):
        # TODO:
        time.sleep(5)
        item.finished.write(False)

    def execute_wait(self, item):
        # get stop time
        tz = dateutil.tz.tzlocal()
        now = datetime.datetime.now(tz)
        if item.operation == "For":
            h, m, s = [float(s) if s != "" else 0.0 for s in item.amount.split(":")]
            total_seconds = 3600.0 * h + 60.0 * m + s
            stop_time = now + datetime.timedelta(0, total_seconds)
        elif item.operation == "Until":
            inputs = {}
            inputs["hour"], inputs["minute"], s = [
                int(s) if s != "" else 0 for s in item.amount.split(":")
            ]
            stop_time = collections.OrderedDict()
            stop_time["seconds"] = s

            def get(current, previous):
                if current in inputs.keys():
                    input = inputs[current]
                else:
                    input = 0
                if stop_time[previous] == 0 or not input == 0:
                    stop_time[current] = input
                else:
                    current_now = getattr(now, current)
                    previous_now = getattr(now, previous)
                    if previous_now <= stop_time[previous]:
                        stop_time[current] = current_now
                    else:
                        stop_time[current] = current_now + 1

            previous = "seconds"
            keys = ["minute", "hour", "day", "month", "year"]
            for key in keys:
                get(key, previous)
                previous = key
            stop_time = datetime.datetime(*stop_time.values()[::-1] + [0, tz])
        # wait until stop time
        total_time = (stop_time - now).total_seconds()
        time_remaining = total_time
        while time_remaining > 0:
            time.sleep(1)
            now = datetime.datetime.now(tz)
            time_remaining = (stop_time - now).total_seconds()
            self.fraction_complete.write((total_time - time_remaining) / total_time)
            # check for pause
            while self.queue_status.pause.read():
                self.queue_status.paused.write(True)
                self.queue_status.pause.wait_for_update()
            if not self.queue_status.go.read():
                return
        # send message on slack
        if g.slack_enabled.read():
            message = ":timer_clock: wait complete - {0}".format(item.description)
            g.slack_control.read().send_message(message)

        item.finished.write(True)

    def execute_interrupt(self, item):
        self.queue_status.go.write(False)
        if g.slack_enabled.read():
            message = (
                ":octagonal_sign: Interrupted - {0}\n{1}\nUse `run` command to continue".format(
                    item.description, item.info
                )
            )
            g.slack_control.read().send_message(message)
        item.finished.write(True)


### queue class ###############################################################


class QueueStatus(QtCore.QObject):
    def __init__(self):
        QtCore.QObject.__init__(self)
        self.go = pc.Busy()
        self.going = pc.Busy()
        self.pause = pc.Busy()
        self.paused = pc.Busy()
        self.stop = pc.Busy()
        self.stopped = pc.Busy()
        self.runtime = 0.0  # seconds
        self.last_started = None
        self.tz = dateutil.tz.tzlocal()

    def run_timer(self):
        self.last_started = datetime.datetime.now(self.tz)

    def stop_timer(self):
        now = datetime.datetime.now(self.tz)
        self.runtime += (now - self.last_started).total_seconds()

    def get_runtime(self):
        """
        returns total seconds run
        """
        out = self.runtime
        if self.last_started is not None:
            now = datetime.datetime.now(self.tz)
            out += (now - self.last_started).total_seconds()
        return out


class Queue:
    def __init__(self, name, gui, folder=None, url=None):
        self.name = name[:10]  # cannot be more than 10 charachters
        self.gui = gui
        self.status = gui.queue_status
        self.timestamp = wt.kit.TimeStamp()
        # create queue folder
        if folder is None:
            folder_name = " ".join([self.timestamp.path, self.name])
            self.folder = pc.Value(os.path.join(g.main_window.read().data_folder, folder_name))
            os.mkdir(self.folder.read())
        else:
            self.folder = pc.Value(folder)
            folder_name = os.path.basename(self.folder.read())
        # create queue file
        self.ini_path = os.path.abspath(os.path.join(self.folder.read(), "queue.ini"))
        if not os.path.isfile(self.ini_path):
            with open(self.ini_path, "a"):
                os.utime(self.ini_path, None)  # quickly create empty file
        self.ini = wt.kit.INI(self.ini_path)  # I don't use ini_handler here
        # parameters and status indicators
        self.items = []
        self.index = pc.Value(0)
        self.going = pc.Busy()
        self.paused = pc.Busy()
        # create storage folder on google drive
        if url is None:
            if g.google_drive_enabled.read():
                g.google_drive_control.read().reserve_id(self.folder.read())
                self.url = g.google_drive_control.read().id_to_open_url(self.folder.read())
                g.google_drive_control.read().create_folder(
                    path=self.folder.read(), id=self.folder.read()
                )
                g.google_drive_control.read().create_file(
                    self.ini_path, self.folder.read(), self.ini_path
                )
            else:
                self.url = None
        else:
            self.url = url
        # initialize worker
        self.worker_enqueued = pc.Enqueued()
        self.worker_busy = pc.Busy()
        self.worker = Worker(
            self.worker_enqueued,
            self.worker_busy,
            self.status,
            self.url,
            self.index,
            self.folder,
        )
        self.worker.fraction_complete.updated.connect(self.update_progress)
        self.worker.action_complete.connect(self.on_action_complete)
        self.worker_thread = QtCore.QThread()
        self.worker.moveToThread(self.worker_thread)
        g.shutdown.read().connect(self.worker_thread.quit)
        self.worker_thread.start()
        self.worker_q = pc.Q(self.worker_enqueued, self.worker_busy, self.worker)
        # message on slack
        if g.slack_enabled.read():
            message = ":baby: new queue created - {0} - {1}".format(folder_name, self.url)
            g.slack_control.read().send_message(message)

    def _start_next_action(self):
        self.status.pause.write(False)
        self.status.paused.write(False)
        self.status.stop.write(False)
        self.status.stopped.write(False)
        self.gui.progress_bar.begin_new_scan_timer()
        item = self.items[self.index.read()]
        item.status = "RUNNING"
        if isinstance(item, Interrupt):
            # This needs to run on the main thread to avoid Seg Fault
            if item.info.strip():
                msg = QtWidgets.QMessageBox(
                    QtWidgets.QMessageBox.NoIcon,
                    f"Interrupt: {item.name}",
                    item.info,
                    parent=self.gui.parent_widget,
                )
                msg.addButton("Resume Queue", QtWidgets.QMessageBox.ActionRole)
                msg.addButton("Dismiss", QtWidgets.QMessageBox.AcceptRole)
                msg.setModal(False)
                msg.buttonClicked.connect(self.resume_queue)
                msg.show()
        self.worker_q.push("excecute", item)
        self.gui.message_widget.setText(item.description.upper())

    def append_acquisition(self, aqn_path, update=True):
        # get properties
        ini = wt.kit.INI(aqn_path)
        module_name = ini.read("info", "module")
        item_name = ini.read("info", "name")
        info = ini.read("info", "info")
        description = ini.read("info", "description")
        module = self.gui.modules[module_name]
        # create item
        acquisition = Acquisition(
            aqn_path, module, name=item_name, info=info, description=description
        )
        # append and update
        self.items.append(acquisition)
        if update:
            self.update()

    def append_device(self):
        # TODO:
        print("append_device")

    def append_hardware(self, hardwares, value, units, name, info, description, update=True):
        hardware = Hardware(hardwares, value, units, name=name, info=info, description=description)
        self.items.append(hardware)
        if update:
            self.update()

    def append_script(self):
        # TODO:
        print("append_script")

    def append_wait(self, operation, amount, name, info, description):
        # create item
        wait = Wait(operation, amount, name=name, info=info, description=description)
        # append and update
        self.items.append(wait)
        self.update()

    def append_interrupt(self, name, info, description):
        # create item
        interrupt = Interrupt(name=name, info=info, description=description)
        # append and update
        self.items.append(interrupt)
        self.update()

    def change_index(self, current_index, new_index):
        item = self.items.pop(current_index)
        if new_index < self.index.read():
            new_index = self.index.read() + 1
        self.items.insert(new_index, item)
        self.update()
        return item

    def exit(self):
        # cleanly exit thread
        self.worker_thread.exit()
        self.worker_thread.quit()

    def get_runtime(self):
        seconds = self.status.get_runtime()
        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)
        string = ":".join([str(int(h)).zfill(3), str(int(m)).zfill(2), str(int(s)).zfill(2)])
        return string

    def interrupt(self, option=None, message="Please choose how to proceed."):
        self.gui.queue_control.set_style("WAITING", "stop")
        # pause
        self.status.pause.write(True)
        while not self.status.paused.read():
            time.sleep(0.01)
        # ask user how to proceed
        options = ["RESUME", "SKIP", "STOP"]
        self.gui.interrupt_choice_window.set_text(message)
        if option is None:
            index_chosen = self.gui.interrupt_choice_window.show()
        else:
            index_chosen = wt.kit.get_index(options, option)
        chosen = options[index_chosen]
        # proceed
        if chosen == "RESUME":
            self.status.pause.write(False)
        elif chosen == "SKIP":
            self.status.stop.write(True)
            self.status.pause.write(False)
        elif chosen == "STOP":
            self.status.stop.write(True)
            self.status.go.write(False)
            self.status.pause.write(False)
        # wait for stop
        if chosen in ["SKIP", "STOP"]:
            while not self.status.stopped.read():
                time.sleep(0.01)
        # finish
        self.status.stop.write(False)
        self.status.pause.write(False)
        self.update()

    def finish(self):
        self.status.stop_timer()
        self.gui.on_queue_finished()
        self.update()
        if g.slack_enabled.read():
            g.slack_control.read().send_message(
                ":bell: queue emptied - total runtime {}".format(self.get_runtime())
            )

    def on_action_complete(self):
        # update current item
        item = self.items[self.index.read()]
        if item.finished.read():
            item.status = "COMPLETE"
        else:
            item.status = "FAILED"
        # onto next item
        self.index.write(self.index.read() + 1)
        queue_done = len(self.items) == self.index.read()
        # check if any more items exist in queue
        if queue_done:
            self.finish()
        # continue (if still going)
        if self.status.go.read() and not queue_done:
            self._start_next_action()
        # finish
        self.update()

    def pop(self, index):
        out = self.items.pop(index)
        self.update()
        return out

    def resume_queue(self, button):
        if not self.status.go.read() and button.text() == "Resume Queue":
            self.run()

    def run(self):
        # status
        self.status.run_timer()
        self.status.go.write(True)
        self.status.pause.write(False)
        # excecute next item
        self._start_next_action()
        # finish
        self.update()

    def update(self):
        # update ini
        self.ini.clear()
        self.ini.add_section("info")
        self.ini.write("info", "Yaqc_cmds version", g.version.read())
        self.ini.write("info", "created", self.timestamp.RFC3339)
        self.ini.write("info", "runtime", self.get_runtime())
        self.ini.write("info", "name", self.name)
        self.ini.write("info", "url", self.url)
        for index, item in enumerate(self.items):
            index_str = str(index).zfill(3)
            self.ini.add_section(index_str)
            self.ini.write(index_str, "type", item.type)
            self.ini.write(index_str, "name", item.name)
            self.ini.write(index_str, "info", item.info)
            self.ini.write(index_str, "description", item.description)
            self.ini.write(index_str, "status", item.status)
            self.ini.write(index_str, "created", item.created.RFC3339)
            if item.started is not None:
                self.ini.write(index_str, "started", item.started.RFC3339)
            if item.exited is not None:
                self.ini.write(index_str, "exited", item.exited.RFC3339)
            # allow item to write additional information
            item.write_to_ini(self.ini, index_str)
            if item.status != "ENQUEUED" and hasattr(item, "module"):
                # KFS: This is a slight bit of a hack to get around editing the queue file from worker thread
                # Rather than replacing the path in the queue when the file is copied, it is done here in update
                path = (
                    os.path.join(
                        self.folder.read(),
                        " ".join([index_str, item.module.module_name, item.name]).rstrip(),
                    )
                    + ".aqn"
                )
                if os.path.exists(path):
                    self.ini.write(index_str, "path", path)
        # update display
        self.gui.update_ui()
        # upload ini
        if g.google_drive_enabled.read():
            g.google_drive_control.read().update_file(self.ini_path, id=self.ini_path)

    def update_progress(self):
        # progress bar
        self.gui.progress_bar.set_fraction(self.worker.fraction_complete.read())
        # queue timer
        runtime_string = self.get_runtime()
        self.gui.runtime.write(runtime_string)


### GUI #######################################################################


class GUI(QtCore.QObject):
    def __init__(self, parent_widget, message_widget):
        QtCore.QObject.__init__(self)
        self.progress_bar = g.progress_bar
        # frame, widgets
        self.message_widget = message_widget
        self.parent_widget = parent_widget
        parent_widget.setLayout(QtWidgets.QHBoxLayout())
        parent_widget.layout().setContentsMargins(0, 10, 0, 0)
        self.layout = parent_widget.layout()
        self.create_frame()
        self.interrupt_choice_window = pw.ChoiceWindow(
            "QUEUE INTERRUPTED", button_labels=["RESUME", "SKIP", "STOP"]
        )
        # queue
        self.queue = None
        self.queue_status = QueueStatus()

    def add_button_to_table(self, i, j, text, color, method):
        # for some reason, my lambda function does not work when called outside
        # of a dedicated method - Blaise 2016-09-14
        button = pw.SetButton(text, color=color)
        button.setProperty("TableRowIndex", i)
        button.clicked.connect(lambda: method(button.property("TableRowIndex")))
        self.table.setCellWidget(i, j, button)
        return button

    def add_index_to_table(self, i, max_value):
        # for some reason, my lambda function does not work when called outside
        # of a dedicated method - Blaise 2016-09-14
        index = QtWidgets.QDoubleSpinBox()
        StyleSheet = "QDoubleSpinBox{color: custom_color; font: 14px;}".replace(
            "custom_color", g.colors_dict.read()["text_light"]
        )
        StyleSheet += "QScrollArea, QWidget{background: custom_color;  border-color: black; border-radius: 0px;}".replace(
            "custom_color", g.colors_dict.read()["background"]
        )
        StyleSheet += "QWidget:disabled{color: custom_color_1; font: 14px; border: 0px solid black; border-radius: 0px;}".replace(
            "custom_color_1", g.colors_dict.read()["text_disabled"]
        ).replace(
            "custom_color_2", g.colors_dict.read()["widget_background"]
        )
        index.setStyleSheet(StyleSheet)
        index.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)
        index.setSingleStep(1)
        index.setDecimals(0)
        index.setMaximum(max_value)
        index.setAlignment(QtCore.Qt.AlignCenter)
        index.setValue(i)
        index.setProperty("TableRowIndex", i)
        index.editingFinished.connect(
            lambda: self.on_index_changed(index.property("TableRowIndex"), int(index.value()))
        )
        self.table.setCellWidget(i, 0, index)
        return index

    def create_acquisition_frame(self):
        frame = QtWidgets.QWidget()
        frame.setLayout(QtWidgets.QVBoxLayout())
        layout = frame.layout()
        layout.setMargin(0)
        layout.setContentsMargins(0, 0, 0, 0)
        # load aqn file
        self.load_aqn_button = pw.SetButton("LOAD FROM FILE")
        self.load_aqn_button.clicked.connect(self.on_load_aqn)
        layout.addWidget(self.load_aqn_button)

        input_table = pw.InputTable()
        # module combobox
        self.module_combobox = pc.Combo()
        input_table.add("Acquisition Module", self.module_combobox)
        # name
        self.acquisition_name = pc.String(max_length=10)
        input_table.add("Name", self.acquisition_name)
        # info
        self.acquisition_info = pc.String()
        input_table.add("Info", self.acquisition_info)
        layout.addWidget(input_table)
        # module container widget
        self.module_container_widget = QtWidgets.QWidget()
        self.module_container_widget.setLayout(QtWidgets.QVBoxLayout())
        module_layout = self.module_container_widget.layout()
        module_layout.setMargin(0)
        module_layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.module_container_widget)
        return frame

    def create_device_frame(self):
        frame = QtWidgets.QWidget()
        frame.setLayout(QtWidgets.QVBoxLayout())
        layout = frame.layout()
        layout.setContentsMargins(0, 0, 0, 0)
        # name and info
        input_table = pw.InputTable()
        self.device_name = pc.String(max_length=10)
        input_table.add("Name", self.device_name)
        self.device_info = pc.String()
        input_table.add("Info", self.device_info)
        layout.addWidget(input_table)
        # not implemented message
        label = QtWidgets.QLabel("device not currently implemented")
        StyleSheet = "QLabel{color: custom_color; font: bold 14px}".replace(
            "custom_color", g.colors_dict.read()["text_light"]
        )
        label.setStyleSheet(StyleSheet)
        layout.addWidget(label)
        return frame

    def create_frame(self):
        # queue display -------------------------------------------------------
        # container widget
        display_container_widget = pw.ExpandingWidget()
        display_layout = display_container_widget.layout()
        display_layout.setMargin(0)
        self.layout.addWidget(display_container_widget)
        # table
        self.table = pw.TableWidget()
        self.table.verticalHeader().hide()
        self.table_cols = collections.OrderedDict()
        self.table_cols["Index"] = 50
        self.table_cols["Type"] = 75
        self.table_cols["Status"] = 85
        self.table_cols["Started"] = 110
        self.table_cols["Exited"] = 110
        self.table_cols["Description"] = 200  # expanding
        self.table_cols["Remove"] = 75
        self.table_cols["Load"] = 75
        for i in range(len(self.table_cols.keys())):
            self.table.insertColumn(i)
        labels = list(self.table_cols.keys())
        labels[-1] = ""
        labels[-2] = ""
        self.table.setHorizontalHeaderLabels(labels)
        self.table.horizontalHeader().setSectionResizeMode(5, QtWidgets.QHeaderView.Stretch)
        self.table.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        for i, width in enumerate(self.table_cols.values()):
            self.table.setColumnWidth(i, width)
        display_layout.addWidget(self.table)
        # line ----------------------------------------------------------------
        line = pw.Line("V")
        self.layout.addWidget(line)
        # controls ------------------------------------------------------------
        settings_container_widget = QtWidgets.QWidget()
        settings_scroll_area = pw.scroll_area()
        settings_scroll_area.setWidget(settings_container_widget)
        settings_scroll_area.setMinimumWidth(300)
        settings_scroll_area.setMaximumWidth(300)
        settings_container_widget.setLayout(QtWidgets.QVBoxLayout())
        settings_layout = settings_container_widget.layout()
        settings_layout.setMargin(5)
        self.layout.addWidget(settings_scroll_area)
        # new queue name
        input_table = pw.InputTable()
        self.new_queue_name = pc.String("queue", max_length=10)
        input_table.add("Create Queue", None)
        input_table.add("New Name", self.new_queue_name)
        settings_layout.addWidget(input_table)
        # new queue button
        self.new_queue_button = pw.SetButton("MAKE NEW QUEUE")
        self.new_queue_button.clicked.connect(self.create_new_queue)
        g.queue_control.disable_when_true(self.new_queue_button)
        settings_layout.addWidget(self.new_queue_button)
        # load button
        self.load_button = pw.SetButton("OPEN QUEUE")
        self.load_button.clicked.connect(self.on_open_queue)
        g.queue_control.disable_when_true(self.load_button)
        settings_layout.addWidget(self.load_button)
        # current queue name
        input_table = pw.InputTable()
        self.queue_name = pc.String(display=True)
        input_table.add("Name", self.queue_name)
        # current queue timestamp
        self.queue_timestamp = pc.String(display=True)
        input_table.add("Timestamp", self.queue_timestamp)
        settings_layout.addWidget(input_table)
        # horizontal line
        line = pw.line("H")
        settings_layout.addWidget(line)
        # adjust queue label
        input_table = pw.InputTable()
        input_table.add("Control Queue", None)
        settings_layout.addWidget(input_table)
        # go button
        self.queue_control = pw.QueueControl()
        self.queue_control.clicked.connect(self.on_queue_control_clicked)
        settings_layout.addWidget(self.queue_control)
        self.queue_control.setDisabled(True)
        # queue runtime
        input_table = pw.InputTable()
        self.runtime = pc.String(initial_value="000:00:00", display=True)
        input_table.add("Queue Runtime", self.runtime)
        settings_layout.addWidget(input_table)
        # horizontal line
        line = pw.Line("H")
        settings_layout.addWidget(line)
        # type combobox
        input_table = pw.InputTable()
        allowed_values = [
            "Acquisition",
            "Wait",
            "Interrupt",
            "Hardware",
            "Device",
            "Script",
        ]
        allowed_values.remove("Device")  # not ready yet
        allowed_values.remove("Script")  # not ready yet
        self.type_combo = pc.Combo(allowed_values=allowed_values)
        self.type_combo.updated.connect(self.update_type)
        input_table.add("Add to Queue", None)
        input_table.add("Type", self.type_combo)
        settings_layout.addWidget(input_table)
        # frames
        self.type_frames = collections.OrderedDict()
        self.type_frames["Acquisition"] = self.create_acquisition_frame()
        self.type_frames["Wait"] = self.create_wait_frame()
        self.type_frames["Interrupt"] = self.create_interrupt_frame()
        self.type_frames["Hardware"] = self.create_hardware_frame()
        self.type_frames["Device"] = self.create_device_frame()
        self.type_frames["Script"] = self.create_script_frame()
        for frame in self.type_frames.values():
            settings_layout.addWidget(frame)
            frame.hide()
        self.update_type()
        # append button
        self.append_button = pw.SetButton("APPEND TO QUEUE")
        self.append_button.setDisabled(True)
        self.append_button.clicked.connect(self.on_append_to_queue)
        settings_layout.addWidget(self.append_button)
        # finish --------------------------------------------------------------
        settings_layout.addStretch(1)

    def create_hardware_frame(self):
        frame = QtWidgets.QWidget()
        frame.setLayout(QtWidgets.QVBoxLayout())
        layout = frame.layout()
        layout.setContentsMargins(0, 0, 0, 0)
        # name and info
        input_table = pw.InputTable()
        self.hardware_name = pc.String(max_length=10)
        input_table.add("Name", self.hardware_name)
        self.hardware_info = pc.String()
        input_table.add("Info", self.hardware_info)
        layout.addWidget(input_table)
        self.hardware_hardwares = {}
        for hw in all_hardwares:
            checkbox = pc.Bool()
            input_table.add(hw.name, checkbox)
            self.hardware_hardwares[hw.name] = checkbox

        self.hardware_value = pc.Number()
        input_table.add("Value", self.hardware_value)
        self.hardware_units = pc.String()
        input_table.add("Units", self.hardware_units)

        return frame

    def create_new_queue(self):
        # exit old queue
        if self.queue is not None:
            self.queue.exit()
        # make new queue
        self.queue = Queue(self.new_queue_name.read(), self)
        self.queue_name.write(self.queue.name)
        self.queue_timestamp.write(self.queue.timestamp.path[-5:])
        self.message_widget.setText("QUEUE NOT YET RUN")
        self.queue.update()  # will call self.update_ui

    def create_script_frame(self):
        frame = QtWidgets.QWidget()
        frame.setLayout(QtWidgets.QVBoxLayout())
        layout = frame.layout()
        layout.setMargin(0)
        layout.setContentsMargins(0, 0, 0, 0)
        input_table = pw.InputTable()
        self.script_name = pc.String(max_length=10)
        input_table.add("Name", self.script_name)
        self.script_info = pc.String()
        input_table.add("Info", self.script_info)
        self.script_path = pc.Filepath()
        input_table.add("Script Path", self.script_path)
        layout.addWidget(input_table)
        return frame

    def create_wait_frame(self):
        frame = QtWidgets.QWidget()
        frame.setLayout(QtWidgets.QVBoxLayout())
        layout = frame.layout()
        layout.setContentsMargins(0, 0, 0, 0)
        input_table = pw.InputTable()
        self.wait_name = pc.String(max_length=10)
        input_table.add("Name", self.wait_name)
        self.wait_info = pc.String()
        input_table.add("Info", self.wait_info)
        allowed_values = ["For", "Until"]
        self.wait_operation = pc.Combo(allowed_values=allowed_values)
        input_table.add("Operation", self.wait_operation)
        self.wait_amount = pc.String()
        input_table.add("Amount (H:M:S)", self.wait_amount)
        layout.addWidget(input_table)
        return frame

    def create_interrupt_frame(self):
        frame = QtWidgets.QWidget()
        frame.setLayout(QtWidgets.QVBoxLayout())
        layout = frame.layout()
        layout.setContentsMargins(0, 0, 0, 0)
        input_table = pw.InputTable()
        self.interrupt_name = pc.String(max_length=10)
        input_table.add("Name", self.interrupt_name)
        self.interrupt_info = pc.String()
        input_table.add("Info", self.interrupt_info)
        layout.addWidget(input_table)
        return frame

    def get_status(self, full=False):
        # called by slack
        make_field = g.slack_control.read().make_field
        make_attachment = g.slack_control.read().make_attachment
        if self.queue is None:
            return ":information_desk_person: no queue has been created", {}
        else:
            message = ":information_desk_person: {}".format(self.message_widget.text())
            time_remaining_str = g.progress_bar.time_remaining.text()
            if not time_remaining_str == "00:00:00":
                message += " - {} remaining".format(time_remaining_str)
            attachments = []
            colors = {
                "ENQUEUED": "#808080",
                "RUNNING": "#FFFF00",
                "COMPLETE": "#00FF00",
                "FAILED": "#FF0000",
            }
            # queue status
            fields = []
            fields.append(make_field("name", self.queue.name, short=True))
            fields.append(make_field("created", self.queue.timestamp.human, short=True))
            fields.append(make_field("runtime", self.queue.get_runtime(), short=True))
            fields.append(
                make_field(
                    "done/total",
                    "/".join([str(self.queue.index.read()), str(len(self.queue.items))]),
                    short=True,
                )
            )
            fields.append(make_field("url", self.queue.url, short=False))
            attachments.append(make_attachment("", fields=fields, color="#00FFFF"))
            # queue items
            for item_index, item in enumerate(self.queue.items):
                if not full and item.status != "RUNNING":
                    continue
                name = " - ".join([str(item_index).zfill(3), item.type, item.description])
                fields = []
                if item.started is not None:
                    fields.append(make_field("started", item.started.human, short=True))
                if item.exited is not None:
                    fields.append(make_field("exited", item.exited.human, short=True))
                if item.type == "acquisition" and item.status in ["COMPLETE", "FAILED"]:
                    fields.append(make_field("url", item.url, short=False))
                attachments.append(
                    make_attachment("", title=name, fields=fields, color=colors[item.status])
                )
            return message, attachments

    def load_modules(self):
        # called by MainWindow.__init__
        g.queue_control.write(False)
        if g.debug.read():
            print("load modules")
        # create acquisition thread
        acquisition_thread = QtCore.QThread()
        g.scan_thread.write(acquisition_thread)
        acquisition_thread.start()
        g.shutdown.read().connect(acquisition_thread.quit)
        # import modules
        # done from modules.ini
        # modules appear in order of import (order of appearance in ini)
        config = yaqc_cmds.__main__.config
        self.modules = {}
        for name, load in config["modules"].items():
            if load:
                path = os.path.join(somatic_folder, "modules", name + ".py")
                module = imp.load_source(name, path)
                if module.load():
                    module.mkGUI()
                    self.modules[module.module_name] = module
                    self.module_container_widget.layout().addWidget(module.gui.frame)
        if len(self.modules) > 0:
            # update module combo
            self.module_combobox.set_allowed_values(list(self.modules.keys()))
            self.module_combobox.updated.connect(self.on_module_combobox_updated)
            self.on_module_combobox_updated()

    def on_append_to_queue(self):
        current_type = self.type_combo.read()
        if current_type == "Acquisition":
            p = self.save_aqn(self.queue.folder.read())
            self.queue.append_acquisition(p)
        elif current_type == "Wait":
            name = self.wait_name.read()
            info = self.wait_info.read()
            operation = self.wait_operation.read()
            amount = self.wait_amount.read()
            description = " ".join(["wait", operation.lower(), amount])
            self.queue.append_wait(
                operation, amount, name=name, info=info, description=description
            )
        elif current_type == "Interrupt":
            name = self.interrupt_name.read()
            info = self.interrupt_info.read()
            description = f"INTERRUPT: {name}"
            self.queue.append_interrupt(name=name, info=info, description=description)
        elif current_type == "Hardware":
            name = self.hardware_name.read()
            info = self.hardware_info.read()
            hardwares = [k for k, v in self.hardware_hardwares.items() if v.read()]
            value = self.hardware_value.read()
            units = self.hardware_units.read()
            description = "%s %f %s" % (hardwares, value, units)
            self.queue.append_hardware(
                hardwares, value, units, name=name, info=info, description=description
            )
        elif current_type == "Device":
            # TODO:
            self.queue.append_device()
        elif current_type == "Script":
            # TODO:
            self.queue.append_script()
        else:
            raise Warning("current_type not recognized in on_append_to_queue")
        # run queue
        if self.queue_status.go.read() and not self.queue_status.going.read():
            self.queue.run()

    def on_queue_control_clicked(self):
        if self.queue_status.going.read():
            self.queue.interrupt()
        else:  # queue not currently running
            self.queue.status.go.write(not self.queue.status.go.read())
            queue_full = len(self.queue.items) == self.queue.index.value
            if not queue_full and self.queue.status.go.read():
                self.queue.run()
        self.update_ui()

    def on_queue_finished(self):
        self.queue_control.set_style("RUN QUEUE", "go")

    def on_index_changed(self, row, new_index):
        if isinstance(row, int):
            index = row
        else:
            index = row.toInt()[0]  # given as QVariant
        self.queue.change_index(index, new_index)

    def on_load_aqn(self):
        # get path from user
        caption = "Choose an aqn file"
        options = "AQN (*.aqn);;All Files (*.*)"
        p = file_dialog_handler.open_dialog(
            caption=caption, directory=str(data_folder), options=options
        )
        # load basic info
        ini = wt.kit.INI(p)
        self.acquisition_name.write(ini.read("info", "name"))
        self.acquisition_info.write(ini.read("info", "info"))
        self.module_combobox.write(ini.read("info", "module"))
        # allow module to load from file
        self.modules[self.module_combobox.read()].gui.load(p)

    def on_load_item(self, row):
        if isinstance(row, int):
            index = row
        else:
            index = row.toInt()[0]  # given as QVariant
        item = self.queue.items[index]
        if item.type == "acquisition":
            self.type_combo.write("Acquisition")
            # load basic info
            p = item.aqn_path
            aqn = wt.kit.INI(p)
            self.acquisition_name.write(aqn.read("info", "name"))
            self.acquisition_info.write(aqn.read("info", "info"))
            self.module_combobox.write(aqn.read("info", "module"))
            # allow module to load from file
            self.modules[self.module_combobox.read()].gui.load(p)
        elif item.type == "device":
            raise NotImplementedError()
        elif item.type == "hardware":
            self.type_combo.write("Hardware")
            self.hardware_name.write(item.name)
            self.hardware_info.write(item.info)
            self.hardware_units.write(item.units)
            self.hardware_value.write(item.value)
            for k, v in self.hardware_hardwares.items():
                for h in item.hardwares:
                    if k == h.name:
                        v.write(True)
                        break
                else:
                    v.write(False)
        elif item.type == "interrupt":
            self.type_combo.write("Wait")
            self.interrupt_name.write(item.name)
            self.interrupt_info.write(item.info)
            self.interrupt_message.write(item.message)
        elif item.type == "script":
            raise NotImplementedError()
        elif item.type == "wait":
            self.type_combo.write("Wait")
            self.wait_name.write(item.name)
            self.wait_info.write(item.info)
            self.wait_operation.write(item.operation)
            self.wait_amount.write(item.amount)
        else:
            raise Exception("item.type not recognized in queue.GUI.on_load_item")

    def on_module_combobox_updated(self):
        for module in self.modules.values():
            module.gui.hide()
        self.modules[self.module_combobox.read()].gui.show()

    def on_open_queue(self):
        # get queue folder
        caption = "Choose Queue directory"
        directory = data_folder
        f = file_dialog_handler.dir_dialog(caption=caption, directory=directory)
        p = os.path.join(f, "queue.ini")
        ini = wt.kit.INI(p)
        # choose operation
        if self.queue is None:
            operation = "REPLACE"
        else:
            # ask user how to proceed
            options = ["APPEND", "REPLACE"]
            choice_window = pw.ChoiceWindow("OPEN QUEUE", button_labels=options)
            index_chosen = choice_window.show()
            operation = options[index_chosen]
        # prepare queue
        if operation == "REPLACE":
            if self.queue is not None:
                self.queue.exit()
            name = ini.read("info", "name")
            url = ini.read("info", "url")
            self.queue = Queue(name, self, folder=f, url=url)
            self.queue.status.run_timer()
            runtime = ini.read("info", "runtime")
            h, m, s = [int(s) for s in runtime.split(":")]
            self.queue.status.runtime = h * 3600 + m * 60 + s
        # append items to queue
        i = 0
        while True:
            section = str(i).zfill(3)
            try:
                item_type = ini.read(section, "type")
            except:
                break
            if item_type == "acquisition":
                p = ini.read(section, "path")
                self.queue.append_acquisition(p, update=False)
            elif item_type == "device":
                raise NotImplementedError
            elif item_type == "hardware":
                name = ini.read(section, "name")
                info = ini.read(section, "info")
                description = ini.read(section, "description")
                hardwares = ini.read(section, "hardwares")
                value = ini.read(section, "value")
                units = ini.read(section, "units")
                self.queue.append_hardware(
                    hardwares,
                    value,
                    units,
                    name=name,
                    info=info,
                    description=description,
                )
            elif item_type == "interrupt":
                name = ini.read(section, "name")
                info = ini.read(section, "info")
                description = ini.read(section, "description")
                self.queue.append_interrupt(name=name, info=info, description=description)
            elif item_type == "script":
                raise NotImplementedError
            elif item_type == "wait":
                name = ini.read(section, "name")
                info = ini.read(section, "info")
                operation = ini.read(section, "operation")
                amount = ini.read(section, "amount")
                description = ini.read(section, "description")
                self.queue.append_wait(
                    operation, amount, name=name, info=info, description=description
                )
            else:
                raise KeyError
            if operation == "REPLACE":
                item = self.queue.items[i]
                status = ini.read(section, "status")
                if status == "RUNNING":
                    item.status = "FAILED"
                    item.started = wt.kit.TimeStamp()
                    item.exited = wt.kit.TimeStamp()
                else:
                    item.status = status
                item.created = wt.kit.timestamp_from_RFC3339(ini.read(section, "created"))
                if ini.has_section("started"):
                    item.started = wt.kit.timestamp_from_RFC3339(ini.read(section, "started"))
                    item.exited = wt.kit.timestamp_from_RFC3339(ini.read(section, "exited"))
                    if item.status == "COMPLETE":
                        item.finished.write(True)
            i += 1
        # manage queue index
        index = -1
        for index, item in enumerate(self.queue.items):
            if item.status == "ENQUEUED":
                break
        else:
            index += 1
        self.queue.index.write(index)
        # finish
        self.queue.update()
        self.queue_name.write(self.queue.name)
        self.queue_timestamp.write(self.queue.timestamp.path[-5:])

    def on_remove_item(self, row):
        if isinstance(row, int):
            index = row
        else:
            index = row.toInt()[0]  # given as QVariant
        self.queue.pop(index)

    def save_aqn(self, folder):
        # all aqn files are first created using this method
        # get filepath
        module_name = self.module_combobox.read()
        name = self.acquisition_name.read()
        timestamp = wt.kit.TimeStamp()
        aqn_name = " ".join([timestamp.path, module_name, name]).rstrip() + ".aqn"
        p = os.path.join(folder, aqn_name)
        # create file
        with open(p, "a"):
            os.utime(p, None)  # quickly create empty file
        # fill with general information
        ini = wt.kit.INI(p)
        ini.add_section("info")
        ini.write("info", "Yaqc_cmds version", g.version.read())
        ini.write("info", "created", timestamp.RFC3339)
        ini.write("info", "module", module_name)
        ini.write("info", "name", name)
        ini.write("info", "info", self.acquisition_info.read())
        ini.write("info", "description", module_name)  # optionally overwritten by module
        # fill with module specific information
        self.modules[module_name].gui.save(p)
        # finish
        return p

    def update_ui(self):
        # buttons -------------------------------------------------------------
        if self.queue:
            queue_go = self.queue_status.go.read()
            queue_going = self.queue_status.going.read()
            # queue control
            self.queue_control.setDisabled(False)
            if queue_go:
                if queue_going:
                    self.queue_control.set_style("INTERRUPT QUEUE", "stop")
                else:
                    self.queue_control.set_style("STOP QUEUE", "stop")
                    self.message_widget.setText("QUEUE WAITING")
            else:
                self.queue_control.set_style("RUN QUEUE", "go")
                self.message_widget.setText("QUEUE STOPPED")
            # append button
            self.append_button.setDisabled(False)
        # table ---------------------------------------------------------------
        # clear table
        for _ in range(self.table.rowCount()):
            self.table.removeRow(0)
        # add elements from queue
        for i, item in enumerate(self.queue.items):
            self.table.insertRow(i)
            # index
            index = self.add_index_to_table(i, len(self.queue.items) - 1)
            if not item.status == "ENQUEUED":
                index.setDisabled(True)
            # type
            label = pw.Label(item.type)
            label.setAlignment(QtCore.Qt.AlignCenter)
            label.setMargin(3)
            self.table.setCellWidget(i, 1, label)
            # status
            label = pw.Label(item.status)
            label.setAlignment(QtCore.Qt.AlignCenter)
            label.setMargin(3)
            self.table.setCellWidget(i, 2, label)
            # started
            if item.started is not None:
                text = item.started.hms
                label = pw.Label(text)
                label.setAlignment(QtCore.Qt.AlignCenter)
                label.setMargin(3)
                self.table.setCellWidget(i, 3, label)
            # exited
            if item.exited is not None:
                text = item.exited.hms
                label = pw.Label(text)
                label.setAlignment(QtCore.Qt.AlignCenter)
                label.setMargin(3)
                self.table.setCellWidget(i, 4, label)
            # description
            label = pw.Label(item.description)
            label.setMargin(3)
            label.setToolTip(item.description)
            self.table.setCellWidget(i, 5, label)
            # remove
            button = self.add_button_to_table(i, 6, "REMOVE", "stop", self.on_remove_item)
            if not item.status == "ENQUEUED":
                button.setDisabled(True)
            # load
            self.add_button_to_table(i, 7, "LOAD", "go", self.on_load_item)

    def update_type(self):
        for frame in self.type_frames.values():
            frame.hide()
        self.type_frames[self.type_combo.read()].show()



#! /usr/bin/env python
### ensure folders exist ######################################################

import matplotlib

matplotlib.use("ps")  # important - images will be generated in worker threads

import sys
from PySide2 import QtWidgets, QtCore

app = QtWidgets.QApplication(sys.argv)

import os
import pathlib


#### import ###################################################################
# BEWARE OF CHANGING ORDER OF IMPORTS!!!!!!!!!

from .project import project_globals as g

g.app.write(app)
g.logger.load()

g.logger.log("info", "Startup", "Yaqc_cmds is attempting startup")

from .project import widgets as pw

from .hardware.hardware import all_initialized

import appdirs
import toml

import yaqc


### define ####################################################################


directory = os.path.abspath(os.path.dirname(__file__))


### version information #######################################################

from .__version__ import __version__

g.version.write(__version__)


### main window ###############################################################


class MainWindow(QtWidgets.QMainWindow):
    shutdown = QtCore.Signal()
    queue_control = QtCore.Signal()

    def __init__(self, config):
        QtWidgets.QMainWindow.__init__(self, parent=None)
        self.config = config
        g.system_name.write(self.config["system_name"])
        g.main_window.write(self)
        g.shutdown.write(self.shutdown)
        self.setWindowTitle("yaqc-cmds %s" % __version__)
        # disable 'x'
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.CustomizeWindowHint)
        self.setWindowFlags(self.windowFlags() & ~QtCore.Qt.WindowCloseButtonHint)
        # set size, position
        self.window_verti_size = 600
        self.window_horiz_size = 1000
        self.setGeometry(0, 0, self.window_horiz_size, self.window_verti_size)
        # self._center()
        self.resize(self.window_horiz_size, self.window_verti_size)
        self._create_main_frame()
        # initialize program
        self._initialize_hardware()
        self._initialize_widgets()
        # open internet things
        self._load_google_drive()
        self._load_witch()
        # populate self
        self.data_folder = pathlib.Path.home() / "yaqc-cmds-data"
        self.data_folder.mkdir(exist_ok=True)
        # somatic system
        from yaqc_cmds.somatic import queue

        self.queue_gui = queue.GUI(self.queue_widget, self.queue_message)
        self.queue_gui.load_modules()
        # log completion
        if g.debug.read():
            print("Yaqc_cmds_ui.MainWindow.__init__ complete")
        g.logger.log("info", "Startup", "Yaqc_cmds MainWindow __init__ finished")
        all_initialized()

    def _create_main_frame(self):
        self.main_frame = QtWidgets.QWidget(parent=self)
        hbox = QtWidgets.QHBoxLayout()
        # hardware ------------------------------------------------------------
        hardware_box = QtWidgets.QVBoxLayout()
        # exit button
        exit_button = pw.Shutdown_button()
        exit_button.setMinimumWidth(300)
        exit_button.setMinimumHeight(30)
        exit_button.shutdown_go.connect(self._shutdown)
        hardware_box.addWidget(exit_button)
        g.queue_control.disable_when_true(exit_button)
        # hardware container widget
        hardware_widget = QtWidgets.QWidget(parent=self.main_frame)
        g.hardware_widget.write(hardware_widget)
        # hardware scroll area
        hardware_scroll_area = pw.scroll_area()
        hardware_scroll_area.setWidget(hardware_widget)
        hardware_box.addWidget(hardware_scroll_area)
        hbox.addLayout(hardware_box)
        # box -----------------------------------------------------------------
        box = QtWidgets.QVBoxLayout()
        box.setMargin(0)
        # module progress bar
        progress_bar = QtWidgets.QProgressBar()
        progress_bar.setTextVisible(False)
        g.progress_bar.write(progress_bar)
        box.addWidget(progress_bar)
        # time elapsed/remaining, queue message
        progress_bar.setLayout(QtWidgets.QHBoxLayout())
        time_elapsed = QtWidgets.QLabel("00:00:00")
        self.queue_message = QtWidgets.QLabel("NO QUEUE")
        time_remaining = QtWidgets.QLabel("00:00:00")
        StyleSheet = "QLabel{color: custom_color; font: bold 14px}".replace(
            "custom_color", g.colors_dict.read()["text_light"]
        )
        time_elapsed.setStyleSheet(StyleSheet)
        self.queue_message.setStyleSheet(StyleSheet)
        time_remaining.setStyleSheet(StyleSheet)
        progress_bar.layout().addWidget(time_elapsed)
        progress_bar.layout().addStretch(1)
        progress_bar.layout().addWidget(self.queue_message)
        progress_bar.layout().addStretch(1)
        progress_bar.layout().addWidget(time_remaining)
        g.progress_bar.give_time_display_elements(time_elapsed, time_remaining)
        # hardware box
        hardware_advanced_widget = QtWidgets.QWidget(parent=self.main_frame)
        hardware_advanced_box = QtWidgets.QVBoxLayout()
        hardware_advanced_box.setContentsMargins(0, 10, 0, 0)
        hardware_advanced_widget.setLayout(hardware_advanced_box)
        g.hardware_advanced_box.write(hardware_advanced_box)
        self.hardware_advanced_box = hardware_advanced_box
        # autonomic box
        coset_widget = QtWidgets.QWidget(parent=self.main_frame)
        g.coset_widget.write(coset_widget)
        # sonomic box
        self.queue_widget = QtWidgets.QWidget(parent=self.main_frame)
        # plot box
        self.plot_widget = QtWidgets.QWidget(parent=self.main_frame)
        # tab widget
        self.tabs = pw.TabWidget()
        self.tabs.addTab(hardware_advanced_widget, "Hardware")
        self.tabs.addTab(coset_widget, "Autonomic")
        self.tabs.addTab(self.queue_widget, "Queue")
        self.tabs.addTab(self.plot_widget, "Plot")
        self.tabs.setCurrentIndex(2)  # start on queue tab
        self.tabs.setContentsMargins(0.0, 0.0, 0.0, 0.0)
        box.addWidget(self.tabs)
        # vertical stretch
        box.addStretch(1)
        hbox.addLayout(box)
        # frame ---------------------------------------------------------------
        hbox.setStretchFactor(box, 100)
        hbox.setGeometry(QtCore.QRect(300, 300, 300, 300))
        self.main_frame.setLayout(hbox)
        self.setCentralWidget(self.main_frame)

    def _initialize_hardware(self):
        if g.debug.read():
            print("initialize hardware")
        # import
        from . import hardware
        from .hardware import opas
        from .hardware import spectrometers
        from .hardware import delays
        from .hardware import filters
        import yaqc_cmds.sensors

    def _initialize_widgets(self):
        if g.debug.read():
            print("initialize widgets")
        # import widgets
        import yaqc_cmds.autonomic.coset
        import yaqc_cmds._plot

    def _load_google_drive(self):
        g.google_drive_enabled.write(self.config.get("google_drive", {}).get("enable", False))
        if g.google_drive_enabled.read():
            g.google_drive_control.write(yaqc.Client(self.config["google_drive"]["port"]))

    def _load_witch(self):
        # check if witch is enabled
        g.slack_enabled.write(self.config.get("slack", {}).get("enable", False))
        if g.slack_enabled.read():
            import yaqc_cmds.project.slack as slack

            # create witch
            self.witch = slack.control
            # begin poll timer
            timer = QtCore.QTimer()
            timer.start(500)  # milliseconds
            self.shutdown.connect(timer.stop)
            g.slack_poll_timer.write(timer)
            g.slack_poll_timer.connect_to_timeout(self.witch.poll)

    def _shutdown(self):
        """
        attempt a clean shutdown
        """
        if g.debug.read():
            print("shutdown")
        g.logger.log("info", "Shutdown", "Yaqc_cmds is attempting shutdown")
        self.shutdown.emit()
        g.shutdown.fire()

    def _center(self):
        # a function which ensures that the window appears in the center of the screen at startup
        screen = QtWidgets.QDesktopWidget().screenGeometry()
        size = self.geometry()
        self.move((screen.width() - size.width()) / 2, (screen.height() - size.height()) / 2)

    def get_status(self, full=False):
        # called by slack
        return self.queue_gui.get_status(full)



import time

from PySide2 import QtWidgets, QtCore

### global classes ############################################################


class SimpleGlobal:
    def __init__(self, initial_value=None):
        self.value = initial_value

    def read(self):
        return self.value

    def write(self, value):
        self.value = value


debug = SimpleGlobal(False)


class PollTimer(SimpleGlobal):
    def connect_to_timeout(self, slot):
        QtWidgets.QAction.connect(self.value, QtCore.SIGNAL("timeout()"), slot)


poll_timer = PollTimer()
slack_poll_timer = PollTimer()


class logger:  # must come before other globals
    def __init__(self):
        pass

    def load(self):
        import yaqc_cmds.project.logging_handler as logging_handler

        self.value = logging_handler.log
        if debug.read():
            self.log("info", "Debug", "Yaqc_cmds is in debug mode")

    def log(self, level, name, message="", origin="name"):
        """
        wrapper of logging method for Yaqc_cmds

        accepts strings

        levels: debug, info, warning, error, critical
        """
        self.value(level, name, message, origin)


logger = logger()

### other globals #############################################################
# alphabetical

app = SimpleGlobal()

colors_dict = SimpleGlobal()

coset_control = SimpleGlobal()

coset_widget = SimpleGlobal()

hardware_advanced_box = SimpleGlobal()

hardware_initialized = SimpleGlobal(False)

google_drive_control = SimpleGlobal()

google_drive_enabled = SimpleGlobal()


class hardware_waits:
    def __init__(self):
        """
        holds value, a list of hardware wait_until_still methods
        """
        self.value = []

    def add(self, method):
        self.value.append(method)

    def give_coset_control(self, control):
        self.coset_control = control

    def wait(self, coset=True):
        if coset:
            self.coset_control.launch()
        for method in self.value:
            method()


hardware_waits = hardware_waits()


class hardware_widget(SimpleGlobal):
    def __init__(self, initial_value=None):
        super().__init__(initial_value)
        self.number_of_widgets = 0

    def write(self, value):
        super().write(value)
        self.value.setLayout(QtWidgets.QVBoxLayout())
        self.value.layout().setMargin(5)
        self.value.layout().addStretch(1)

    def add_to(self, widget):
        self.value.layout().takeAt(self.number_of_widgets)
        self.value.layout().addWidget(widget)
        self.number_of_widgets += 1
        self.value.layout().addStretch(1)


hardware_widget = hardware_widget()

main_thread = SimpleGlobal(QtCore.QThread.currentThread())

main_window = SimpleGlobal()

scan_thread = SimpleGlobal()


class QueueControl(QtCore.QObject):
    def __init__(self):
        self.value = None
        self.widgets_to_disable = []

    def read(self):
        return self.value

    def write(self, value):
        for widget in self.widgets_to_disable:
            try:
                widget.setDisabled(value)
            except RuntimeError:
                # widget has been deleted, probably
                self.widgets_to_disable.remove(widget)
        self.value = value
        main_window.read().queue_control.emit()

    def disable_when_true(self, widget):
        self.widgets_to_disable.append(widget)


queue_control = QueueControl()


class progress_bar:
    def __init__(self):
        self.value = None

    def write(self, value):
        self.value = value

    def give_time_display_elements(self, time_elapsed, time_remaining):
        self.time_elapsed = time_elapsed
        self.time_remaining = time_remaining

    def begin_new_scan_timer(self):
        self.start_time = time.time()

    def set_fraction(self, fraction):
        self.value.setValue(fraction * 100)
        # time elapsed
        time_elapsed = time.time() - self.start_time
        m, s = divmod(time_elapsed, 60)
        h, m = divmod(m, 60)
        self.time_elapsed.setText("%02d:%02d:%02d" % (h, m, s))
        # time remaining
        if fraction == 0:
            self.time_remaining.setText("??:??:??")
        else:
            time_remaining = (time_elapsed / fraction) - time_elapsed
            m, s = divmod(time_remaining, 60)
            h, m = divmod(m, 60)
            self.time_remaining.setText("%02d:%02d:%02d" % (h, m, s))


progress_bar = progress_bar()


class shutdown(SimpleGlobal):
    """
    holds the reference of MainWindow.shutdown Qt signal

    during startup, add your shutdown method to this object using the 'add_method' method it will be called upon shutdown.
    your method must not have any arguments
    """

    def __init__(self, initial_value=None):
        super().__init__(initial_value)
        self.methods = []

    def add_method(self, method):
        self.methods.append(method)

    def fire(self):
        for method in self.methods:
            method()
        main_window.read().close()


shutdown = shutdown()

slack_control = SimpleGlobal()

slack_enabled = SimpleGlobal()

system_name = SimpleGlobal()

version = SimpleGlobal()



import configparser
import pathlib

import numpy as np
import WrightTools as wt

import yaqc_cmds
import somatic.acquisition as acquisition
import yaqc_cmds.project.widgets as pw
import yaqc_cmds.project.classes as pc
import yaqc_cmds.hardware.opas as opas
import yaqc_cmds.hardware.spectrometers as spectrometers
import yaqc_cmds.devices.devices as devices


class Worker(acquisition.Worker):
    def process(self, scan_folder):
        config = self.config_dictionary
        data_path = wt.kit.glob_handler(".data", folder=str(scan_folder))[0]
        data = wt.data.from_Yaqc_cmds(data_path)
        kwargs = {k.lower(): v for k, v in config["Processing"].items()}
        apply_ = kwargs.pop("apply")
        old_path = self.curve.save(scan_folder, plot=False)
        old_path.rename(old_path.with_suffix(f".old{old_path.suffix}"))
        curve = self._process(data, self.curve, config=config, scan_folder=scan_folder, **kwargs)
        if apply_ and not self.stopped.read():
            path = curve.save(pathlib.Path(self.opa_hardware.curve_paths[self.curve_id]).parent)
            self.opa_hardware.driver.curve_paths[self.curve_id].write(str(path))
            yaqc_cmds.somatic.updated_attune_store.emit()
        self.upload(
            scan_folder,
            reference_image=str(pathlib.Path(scan_folder) / self.reference_image),
        )

    def _process(self, data, curve, channel, gtol, ltol, level, scan_folder, config):
        ...

    @property
    def config_dictionary(self):
        out = {}
        for section in self.aqn.sections:
            out[section] = {k: self.aqn.read(section, k) for k in self.aqn.get_options(section)}
        return out

    def run(self):
        full_shape = []
        axes = []

        config = self.config_dictionary
        opa_name = config["OPA"]["opa"]
        opa_names = [opa.name for opa in opas.hardwares]
        opa_index = opa_names.index(opa_name)
        self.opa_hardware = opas.hardwares[opa_index]

        self.curve = self.opa_hardware.curve.copy()
        curve_ids = list(self.opa_hardware.driver.curve_paths.keys())
        for section, conf in config.items():
            if not section.startswith("Motor"):
                continue
            name = conf["motor"]
            break
        else:
            # KFS 2019-10-28: Only applys to tune test in current tuning functions
            # Do not want it using poynting in this case
            if self.curve.kind == "poynting":
                self.curve = self.curve.subcurve
            name = self.curve.dependent_names[0]
        while not name in self.curve.dependent_names:
            self.curve = self.curve.subcurve
            curve_ids = curve_ids[:-1]
        self.curve_id = curve_ids[-1]
        self.curve.convert("wn")

        spec_name = config["Spectrometer"]["spectrometer"]
        spec_action = config["Spectrometer"]["action"]
        spec_names = [spec.name for spec in spectrometers.hardwares]
        spec_index = spec_names.index(spec_name)
        self.spec_hardware = spectrometers.hardwares[spec_index]

        if spec_action == "Zero Order":
            self.spec_hardware.set_position(0)

        axis_identity = opa_name
        if spec_action == "Tracking":
            axis_identity = f"{opa_name}={spec_name}"
        axes.append(acquisition.Axis(self.curve.setpoints[:], "wn", axis_identity, axis_identity))
        full_shape.append(len(self.curve.setpoints))
        for section, conf in config.items():
            if not section.startswith("Motor"):
                continue
            if "num" not in conf:
                continue
            full_shape.append(int(conf["num"]))
        if spec_action == "Scanned":
            for section, conf in config.items():
                if not section.startswith("Spectral"):
                    continue
                full_shape.append(int(conf["num"]))
        print(f"{full_shape}")
        index = 1

        for section, conf in config.items():
            if not section.startswith("Motor"):
                continue
            if "num" not in conf:
                continue
            name = conf["motor"]
            width = conf["width"]
            npts = int(conf["num"])
            points = np.linspace(-width / 2.0, width / 2.0, npts)
            sh = [1] * (len(full_shape) - 1)
            sh[0] = len(self.curve[name])
            centers = self.curve[name][:].reshape(sh)
            print(f"{sh}")
            centers = np.broadcast_to(centers, full_shape[:index] + full_shape[index + 1 :])
            hardware_dict = {opa_name: [self.opa_hardware, "set_motor", [name, "destination"]]}
            axes.append(
                acquisition.Axis(
                    points,
                    None,
                    f"{opa_name}_{name}",
                    f"D{opa_name}",
                    hardware_dict,
                    centers=centers,
                )
            )
            index += 1

        if spec_action == "Scanned":
            for section, conf in config.items():
                if not section.startswith("Spectral"):
                    continue
                name = conf["axis"]
                width = conf["width"]
                npts = int(conf["num"])
                points = np.linspace(-width / 2.0, width / 2.0, npts)
                sh = [1] * (len(full_shape) - 1)
                sh[0] = len(self.curve.setpoints)
                centers = self.curve.setpoints[:].reshape(sh)
                centers = np.broadcast_to(centers, full_shape[:index] + full_shape[index + 1 :])
                axes.append(acquisition.Axis(points, "wn", f"{name}", f"D{name}", centers=centers))
                index += 1

        self.scan(axes)

        if not self.stopped.read():
            self.finished.write(True)


class GUI(acquisition.GUI):
    def __init__(self, module_name):
        self.items.update(
            {
                "OPA": OpaSectionWidget("OPA", self),
                "Spectrometer": SpectrometerSectionWidget("Spectrometer", self),
                "Processing": ProcessingSectionWidget("Processing", self),
            }
        )
        super().__init__(module_name)
        for item in self.items.values():
            for i in item.items.values():
                i.updated.connect(self.on_update)
        self.items["Device Settings"] = self.device_widget
        self.on_update()

    def create_frame(self):
        self.layout.addWidget(self["OPA"].input_table)
        self.layout.addWidget(self["Spectrometer"].input_table)
        for name, item in self.items.items():
            if name in ("OPA", "Spectrometer", "Processing", "Device Settings"):
                continue
            self.layout.addWidget(item.input_table)
        self.layout.addWidget(self["Processing"].input_table)

    def load(self, aqn_path):
        for item in self.items.values():
            item.load(aqn_path)

    def save(self, aqn_path):
        aqn = wt.kit.INI(aqn_path)
        aqn.write("info", "description", f"{self['OPA']['OPA'].read()} {self.module_name}")
        for item in self.items.values():
            item.save(aqn_path)

    def on_device_settings_updated(self):
        self.on_update()

    def on_update(self):
        for item in self.items.values():
            if isinstance(item, AqnSectionWidget):
                item.on_update()

    def __getitem__(self, name):
        return self.items[name]


class AqnSectionWidget:
    def __init__(self, section_name, parent):
        self.section_name = section_name
        self.parent = parent
        self.items = {}

    @property
    def input_table(self):
        input_table = pw.InputTable()
        input_table.add(self.section_name, None)
        for name, item in self.items.items():
            input_table.add(name, item)
        return input_table

    def load(self, aqn_path):
        aqn = wt.kit.INI(aqn_path)
        for name, item in self.items.items():
            item.write(aqn.read(self.section_name, name))

    def save(self, aqn_path):
        aqn = wt.kit.INI(aqn_path)
        try:
            aqn.add_section(self.section_name)
        except configparser.DuplicateSectionError:
            pass
        for name, item in self.items.items():
            aqn.write(self.section_name, name, item.read())

    def on_update(self):
        pass

    def __getitem__(self, name):
        return self.items[name]


class OpaSectionWidget(AqnSectionWidget):
    def __init__(self, section_name, parent):
        super().__init__(section_name, parent)
        allowed = [hardware.name for hardware in opas.hardwares]
        self.items["OPA"] = pc.Combo(allowed)


class SpectrometerSectionWidget(AqnSectionWidget):
    def __init__(self, section_name, parent):
        super().__init__(section_name, parent)
        allowed = [hardware.name for hardware in spectrometers.hardwares]
        self.items["Action"] = pc.Combo(
            ["Scanned", "None", "Tracking", "Zero Order"], initial_value="Scanned"
        )
        self.items["Spectrometer"] = pc.Combo(allowed)

    def on_update(self):
        self.items["Spectrometer"].set_disabled(self.items["Action"].read() == "None")


class ProcessingSectionWidget(AqnSectionWidget):
    def __init__(self, section_name, parent):
        super().__init__(section_name, parent)
        self.items["Channel"] = pc.Combo(devices.control.channel_names)
        self.items["level"] = pc.Bool(initial_value=False)
        self.items["gtol"] = pc.Number(initial_value=1e-3, decimals=5)
        self.items["ltol"] = pc.Number(initial_value=1e-2, decimals=5)
        self.items["Apply"] = pc.Bool(initial_value=False)

    def on_update(self):
        self.items["Channel"].set_allowed_values(devices.control.channel_names)


class SpectralAxisSectionWidget(AqnSectionWidget):
    def __init__(self, section_name, parent):
        super().__init__(section_name, parent)
        self.items["Axis"] = pc.Combo()
        self.items["Width"] = pc.Number(initial_value=-250)
        self.items["Num"] = pc.Number(initial_value=51, decimals=0)

    def on_update(self):
        if self.parent["Spectrometer"]["Action"].read() == "Scanned":
            allowed = [self.parent["Spectrometer"]["Spectrometer"].read()]
            for value in self.items.values():
                value.set_disabled(False)
            self["Axis"].set_disabled(True)
        else:
            for value in self.items.values():
                value.set_disabled(True)
            allowed = []
            for dev in devices.control.devices:
                if dev.has_map:
                    allowed.extend(dev.map_axes.keys())
            if not allowed:
                allowed = [""]
        self.items["Axis"].set_allowed_values(allowed)
        self.items["Axis"].set_disabled(len(allowed) > 1)


class MotorAxisSectionWidget(AqnSectionWidget):
    def __init__(self, section_name, parent):
        super().__init__(section_name, parent)
        self.items["Motor"] = pc.Combo()
        self.items["Width"] = pc.Number(initial_value=1, decimals=3)
        self.items["Num"] = pc.Number(initial_value=31, decimals=0)

    def on_update(self):
        hardware = next(h for h in opas.hardwares if h.name == self.parent["OPA"]["OPA"].read())
        self.items["Motor"].set_allowed_values(hardware.curve.dependent_names)


def load():
    return False



#!/usr/bin/env python3

import os
import pathlib
import subprocess
import sys

import appdirs  # type: ignore
import click
import toml

from .__version__ import __version__

config = {}


@click.group(invoke_without_command=True)
@click.version_option(__version__)
@click.pass_context
def main(ctx):
    if ctx.invoked_subcommand is None:
        launch()


@main.command()
@click.option("-c", "--config", "config_filepath")
def launch(config_filepath):
    from ._main_window import app, MainWindow
    from .project import style

    if config_filepath:
        config_filepath = pathlib.Path(config_filepath)
    else:
        config_filepath = (
            pathlib.Path(appdirs.user_config_dir("yaqc-cmds", "yaqc-cmds")) / "config.toml"
        )

    global window, config
    config = toml.load(config_filepath)
    window = MainWindow(config)
    style.set_style()
    window.show()
    window.showMaximized()
    app.exec_()


@main.command(name="edit-config")
def edit_config():
    config_filepath = (
        pathlib.Path(appdirs.user_config_dir("yaqc-cmds", "yaqc-cmds")) / "config.toml"
    )
    config_filepath.parent.mkdir(parents=True, exist_ok=True)
    if not config_filepath.exists():
        if click.confirm(
            "No config file found, would you like to use the default (virtual) configuration?",
            default=True,
        ):
            config_template = pathlib.Path(__file__).parent / "config-template.toml"
            config_filepath.write_text(config_template.read_text())

    while True:
        if sys.platform.startswith("win32"):
            config_filepath = str(config_filepath)
            subprocess.run([os.environ.get("EDITOR", "notepad.exe"), config_filepath])
        else:
            subprocess.run([os.environ.get("EDITOR", "vi"), config_filepath])
        try:
            toml.load(config_filepath)
            break
        except Exception as e:
            print(e, file=sys.stderr)

            if not click.confirm(
                "Error parsing config toml. Would you like to re-edit?",
                default=True,
            ):
                break


if __name__ == "__main__":
    main()



## import ####################################################################


import time
import pathlib

import appdirs
import toml

from PySide2 import QtWidgets

import WrightTools as wt
import attune
import yaqc

import yaqc_cmds.__main__
import yaqc_cmds.project.project_globals as g
import yaqc_cmds.project.widgets as pw
import yaqc_cmds.project.classes as pc
from yaqc_cmds.hardware import hardware as hw
from yaqc_cmds.somatic import signals


### driver ####################################################################


class Driver(hw.Driver):
    def __init__(self, *args, **kwargs):
        self.yaqd_port = kwargs["yaqd_port"]
        self.yaqd_host = kwargs.get("yaqd_host", "127.0.0.1")
        self.client = yaqc.Client(self.yaqd_port, host=self.yaqd_host)
        self.client.register_connection_callback(signals.updated_attune_store.emit)
        signals.updated_attune_store.connect(self.on_updated_attune_store)
        self.motor_positions = {
            k: pc.Number(name=k, decimals=6, display=True) for k in self.client.get_setable_names()
        }
        hw.Driver.__init__(self, *args, **kwargs)
        self.shutter_port = kwargs.get("shutter_yaqd_port")
        if self.shutter_port:
            self.shutter_position = pc.Bool(name="Shutter", display=True, set_method="set_shutter")
            self.shutter = yaqc.Client(self.shutter_port)
            self.shutter.set_identifier("closed")
            self.exposed += [self.shutter_position]
        self.on_updated_attune_store()

    def on_updated_attune_store(self):
        self.curve = attune.Instrument(**self.client.get_instrument())
        self.load_curve()
        self.get_motor_positions()
        self.get_position()

    def is_busy(self):
        return self.client.busy()

    def get_position(self):
        position = self.client.get_position()
        self.position.write(position, self.native_units)
        return position

    def get_motor_positions(self):
        positions = self.client.get_setable_positions()
        for k, v in self.motor_positions.items():
            v.write(positions[k])

    def home_all(self, inputs=None):
        if self.shutter_port:
            self.set_shutter([0])
        self.client.home()

    def home_motor(self, inputs):
        if self.shutter_port:
            self.set_shutter([0])
        self.client.home_setables([inputs])

    def load_curve(self, path=None):
        self.recorded[self.name] = [
            self.position,
            self.native_units,
            None,
            self.label.read(),
            None,
        ]
        for name, sa in self.motor_positions.items():
            self.recorded[f"{self.name}_{name}"] = [sa, None, None, name.lower(), None]
        self.set_arrangement(self.arrangement)

    def set_motor(self, motor_name, destination, wait=True):
        self.client.set_setable_positions({motor_name: destination})
        if wait:
            self.wait_until_still()

    def set_motors(self, motor_names, motor_positions, wait=True):
        destinations = {n: p for n, p in zip(motor_names, motor_positions)}
        self.client.set_setable_positions(destinations)
        if wait:
            self.wait_until_still()

    def set_position(self, destination, units=None):
        if units:
            destination = wt.units.convert(destination, units, self.native_units)
        self.client.set_position(destination)
        self.wait_until_still()
        self.get_position()
        self.save_status()

    def set_position_except(self, destination, exceptions, units=None):
        """
        set position, except for motors that follow

        does not wait until still...
        """
        if units:
            destination = wt.units.convert(destination, units, self.native_units)
        self.hardware.destination.write(destination, self.native_units)
        self.position.write(destination, self.native_units)
        self.client.set_position_except(destination, exceptions)
        self.get_position()
        self.save_status()

    def wait_until_still(self):
        while self.is_busy():
            time.sleep(0.01)
            self.get_motor_positions()
            self.get_position()
        self.get_motor_positions()
        self.get_position()

    def set_arrangement(self, arrangement):
        self.client.set_arrangement(arrangement)
        self.limits.write(*self.client.get_limits(), self.native_units)

    @property
    def arrangement(self):
        return self.client.get_arrangement()

    def get_all_arrangements(self):
        return self.client.get_all_arrangements()

    def set_shutter(self, inputs):
        shutter_state = inputs[0]
        error = self.shutter.set_position(shutter_state)
        self.shutter_position.write(shutter_state)
        return error

    def close(self):
        if self.shutter_port:
            self.set_shutter([0])


### gui #######################################################################


class GUI(hw.GUI):
    def initialize(self):
        arr = self.driver.arrangement
        # self.hardware.driver.initialize()
        # container widget
        display_container_widget = QtWidgets.QWidget()
        display_container_widget.setLayout(QtWidgets.QVBoxLayout())
        display_layout = display_container_widget.layout()
        display_layout.setMargin(0)
        self.layout.addWidget(display_container_widget)
        # plot
        self.plot_widget = pw.Plot1D()
        self.plot_widget.plot_object.setMouseEnabled(False, False)
        self.plot_curve = self.plot_widget.add_scatter()
        self.plot_h_line = self.plot_widget.add_infinite_line(angle=0, hide=False)
        self.plot_v_line = self.plot_widget.add_infinite_line(angle=90, hide=False)
        display_layout.addWidget(self.plot_widget)
        # vertical line
        line = pw.line("V")
        self.layout.addWidget(line)
        # container widget / scroll area
        settings_container_widget = QtWidgets.QWidget()
        settings_scroll_area = pw.scroll_area()
        settings_scroll_area.setWidget(settings_container_widget)
        settings_scroll_area.setMinimumWidth(300)
        settings_scroll_area.setMaximumWidth(300)
        settings_container_widget.setLayout(QtWidgets.QVBoxLayout())
        settings_layout = settings_container_widget.layout()
        settings_layout.setMargin(5)
        self.layout.addWidget(settings_scroll_area)
        # opa properties
        input_table = pw.InputTable()
        settings_layout.addWidget(input_table)
        # plot control
        input_table = pw.InputTable()
        input_table.add("Display", None)
        self.plot_motor = pc.Combo(allowed_values=self.driver.curve.arrangements[arr].keys())
        self.plot_motor.updated.connect(self.update_plot)
        input_table.add("Motor", self.plot_motor)
        allowed_values = list(wt.units.get_valid_conversions("wn"))
        self.plot_units = pc.Combo(
            initial_value=self.driver.native_units, allowed_values=allowed_values
        )
        self.plot_units.updated.connect(self.update_plot)
        input_table.add("Units", self.plot_units)
        settings_layout.addWidget(input_table)
        # curves
        input_table = pw.InputTable()
        # input_table.add("Curves", None)
        self.arrangement_combo = pc.Combo(allowed_values=self.driver.curve.arrangements.keys())
        self.arrangement_combo.write(arr)
        self.arrangement_combo.updated.connect(self.on_arrangement_updated)
        input_table.add("Arrangement", self.arrangement_combo)
        # limits
        limits = pc.NumberLimits()  # units None
        self.low_energy_limit_display = pc.Number(
            units=self.driver.native_units, display=True, limits=limits
        )
        input_table.add("Low Energy Limit", self.low_energy_limit_display)
        self.high_energy_limit_display = pc.Number(
            units=self.driver.native_units, display=True, limits=limits
        )
        input_table.add("High Energy Limit", self.high_energy_limit_display)
        settings_layout.addWidget(input_table)
        self.driver.limits.updated.connect(self.on_limits_updated)
        # motors
        input_table = pw.InputTable()
        input_table.add("Setable", None)
        settings_layout.addWidget(input_table)
        for motor_name, motor_mutex in self.driver.motor_positions.items():
            settings_layout.addWidget(MotorControlGUI(motor_name, motor_mutex, self.driver))
        self.home_all_button = pw.SetButton("HOME ALL", "advanced")
        settings_layout.addWidget(self.home_all_button)
        self.home_all_button.clicked.connect(self.on_home_all)
        g.queue_control.disable_when_true(self.home_all_button)
        # stretch
        settings_layout.addStretch(1)
        # signals and slots
        self.arrangement_combo.updated.connect(self.update_plot)
        self.driver.update_ui.connect(self.update)
        # finish
        self.update()
        self.update_plot()
        self.on_limits_updated()
        signals.updated_attune_store.connect(self.update_plot)

    def update(self):
        # set button disable
        if self.driver.busy.read():
            self.home_all_button.setDisabled(True)
            for motor_mutex in self.driver.motor_positions.values():
                motor_mutex.set_disabled(True)
        else:
            self.home_all_button.setDisabled(False)
            for motor_mutex in self.driver.motor_positions.values():
                motor_mutex.set_disabled(False)
        # update destination motor positions
        # TODO:
        # update plot lines
        motor_name = self.plot_motor.read()
        try:
            motor_position = self.driver.motor_positions[motor_name].read()
            self.plot_h_line.setValue(motor_position)
        except:
            pass
        units = self.plot_units.read()
        self.plot_v_line.setValue(self.driver.position.read(units))

    def update_plot(self):
        arr = self.arrangement_combo.read()
        motor_name = self.plot_motor.read()
        tune = self.driver.curve.arrangements[arr][motor_name]
        if not isinstance(tune, attune.Tune):
            return
        # units
        units = self.plot_units.read()
        # xi
        colors = tune.independent
        xi = wt.units.converter(colors, tune.ind_units, units)
        # yi
        yi = tune.dependent
        self.plot_widget.set_labels(xlabel=units, ylabel=motor_name)
        self.plot_curve.clear()
        try:
            self.plot_curve.setData(xi, yi)
        except ValueError:
            pass
        self.plot_widget.graphics_layout.update()
        self.update()

    def on_curve_paths_updated(self):
        self.driver.load_curve()  # TODO: better
        self.update_plot()

    def on_home_all(self):
        self.hardware.q.push("home_all")

    def on_limits_updated(self):
        low_energy_limit, high_energy_limit = self.driver.limits.read("wn")
        self.low_energy_limit_display.write(low_energy_limit, "wn")
        self.high_energy_limit_display.write(high_energy_limit, "wn")

    def on_arrangement_updated(self):
        arr = self.arrangement_combo.read()
        self.driver.set_arrangement(arr)
        self.plot_motor.set_allowed_values(self.driver.curve.arrangements[arr].keys())
        self.update_plot()


class MotorControlGUI(QtWidgets.QWidget):
    def __init__(self, motor_name, motor_mutex, driver):
        QtWidgets.QWidget.__init__(self)
        self.motor_name = motor_name
        self.driver = driver
        self.hardware = driver.hardware
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.setMargin(0)
        # table
        input_table = pw.InputTable()
        input_table.add(motor_name, motor_mutex)
        self.destination = motor_mutex.associate(display=False)
        input_table.add("Dest. " + motor_name, self.destination)
        self.layout.addWidget(input_table)
        # buttons
        home_button, set_button = self.add_buttons(self.layout, "HOME", "advanced", "SET", "set")
        home_button.clicked.connect(self.on_home)
        set_button.clicked.connect(self.on_set)
        g.queue_control.disable_when_true(home_button)
        g.queue_control.disable_when_true(set_button)
        # finish
        self.setLayout(self.layout)

    def add_buttons(self, layout, button1_text, button1_color, button2_text, button2_color):
        colors = g.colors_dict.read()
        # layout
        button_container = QtWidgets.QWidget()
        button_container.setLayout(QtWidgets.QHBoxLayout())
        button_container.layout().setMargin(0)
        # button1
        button1 = QtWidgets.QPushButton()
        button1.setText(button1_text)
        button1.setMinimumHeight(25)
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors[button1_color]
        )
        button1.setStyleSheet(StyleSheet)
        button_container.layout().addWidget(button1)
        g.queue_control.disable_when_true(button1)
        # button2
        button2 = QtWidgets.QPushButton()
        button2.setText(button2_text)
        button2.setMinimumHeight(25)
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors[button2_color]
        )
        button2.setStyleSheet(StyleSheet)
        button_container.layout().addWidget(button2)
        g.queue_control.disable_when_true(button2)
        # finish
        layout.addWidget(button_container)
        return [button1, button2]

    def on_home(self):
        self.hardware.home_motor(self.motor_name)

    def on_set(self):
        destination = self.destination.read()
        self.hardware.set_motor(self.motor_name, destination)


### hardware ##################################################################


class Hardware(hw.Hardware):
    def __init__(self, *arks, **kwargs):
        self.kind = "OPA"
        hw.Hardware.__init__(self, *arks, **kwargs)

    @property
    def curve(self):
        return self.driver.curve

    def home_motor(self, motor):
        self.q.push("home_motor", motor)

    def load_curve(self, name, path):
        self.q.push("load_curve", name, path)

    @property
    def motor_names(self):
        return list(self.driver.motor_positions.keys())

    def set_motor(self, motor, destination):
        self.q.push("set_motor", motor, destination)

    @property
    def arrangement(self):
        return self.driver.arrangement


### initialize ################################################################


conf = yaqc_cmds.__main__.config
hardwares, gui, advanced_gui = hw.import_hardwares(
    conf.get("hardware", {}).get("opas", {}),
    name="OPAs",
    Driver=Driver,
    GUI=GUI,
    Hardware=Hardware,
)



### import ####################################################################

import pathlib
import time

import numpy as np

import matplotlib

matplotlib.pyplot.ioff()

from PySide2 import QtWidgets
import WrightTools as wt

import yaqc_cmds.project.project_globals as g
import yaqc_cmds.project.classes as pc
import yaqc_cmds.project.widgets as pw
import yaqc_cmds.somatic.acquisition as acquisition

import yaqc_cmds
import yaqc_cmds.hardware.spectrometers as spectrometers
import yaqc_cmds.hardware.delays as delays
import yaqc_cmds.hardware.opas as opas
import yaqc_cmds.hardware.filters as filters
from yaqc_cmds.somatic import _wt5

all_hardwares = opas.hardwares + spectrometers.hardwares + delays.hardwares + filters.hardwares


### define ####################################################################


module_name = "SCAN"


### custom classes ############################################################


class Axis:
    def __init__(self, units, axis_index):
        self.units = units
        # TODO better defaults system...
        if self.units == "wn":
            initial_start = 1500
            initial_stop = 1200
        elif self.units == "ps":
            initial_start = -1
            initial_stop = 1
        elif self.units == "deg":
            initial_start = 0.0
            initial_stop = 360.0
        else:
            raise ValueError(f"unexpected units: {self.units}")
        self.widget = pw.InputTable()
        self.widget.add(str(axis_index) + " (" + self.units + ")", None)
        # start
        self.start = pc.Number(initial_value=initial_start, units=self.units)
        self.start.set_disabled_units(True)
        self.widget.add("Initial", self.start)
        # stop
        self.stop = pc.Number(initial_value=initial_stop, units=self.units)
        self.stop.set_disabled_units(True)
        self.widget.add("Final", self.stop)
        # number
        self.number = pc.Number(initial_value=51, decimals=0)
        self.widget.add("Number", self.number)
        # hardwares
        if self.units == "wn":
            hardware_objs = opas.hardwares + spectrometers.hardwares
        elif self.units == "ps":
            hardware_objs = delays.hardwares
        elif self.units == "deg":
            hardware_objs = filters.hardwares
        self.hardwares = {}
        for hw in hardware_objs:
            checkbox = pc.Bool()
            self.widget.add(hw.name, checkbox)
            self.hardwares[hw.name] = checkbox

    def get_name(self):
        return "=".join([key for key in self.hardwares if self.hardwares[key].read()])

    def hide(self):
        self.widget.hide()


class Constant:
    def __init__(self):
        self.widget = pw.InputTable()
        self.widget.add("Constant", None)
        # hardware name
        allowed_values = [h.name for h in all_hardwares]
        self.hardware_name_combo = pc.Combo(allowed_values=allowed_values)
        self.hardware_name_combo.write(spectrometers.hardwares[0].name)
        # self.hardware_name_combo.set_disabled(True)
        self.widget.add("Hardware", self.hardware_name_combo)
        # expression
        opanames = [h.name for h in opas.hardwares]
        self.expression = pc.String(initial_value="+".join(opanames))
        self.widget.add("Expression", self.expression)

    def get_name(self):
        return self.hardware_name_combo.read()

    def hide(self):
        self.widget.hide()


### Worker ####################################################################


class Worker(acquisition.Worker):
    def process(self, scan_folder):
        with _wt5.data_container as data:
            # decide which channels to make plots for
            main_channel = self.aqn.read("processing", "main channel")
            if self.aqn.read("processing", "process all channels"):
                channels = data.channel_names
            else:
                channels = [main_channel]
            # make figures for each channel
            data_path = pathlib.Path(_wt5.data_container.data_filepath)
            data_folder = data_path.parent
            # make all images
            for channel_name in channels:
                channel_path = data_folder / channel_name
                output_path = data_folder
                orig_transform = data.axis_names
                print(data.axis_names)
                axes = []
                for a in orig_transform:
                    print(data[channel_name].shape, data[a].shape)
                    if all(
                        ch_sh >= a_sh
                        for ch_sh, a_sh in zip(data[channel_name].shape, data[a].shape)
                    ):
                        axes.append(a)
                data.transform(*axes)
                print(data.axis_names)
                if sum(a > 1 for a in data[channel_name].shape) > 2:
                    output_path = channel_path
                    channel_path.mkdir()
                channel_index = data.channel_names.index(channel_name)
                image_fname = channel_name
                if sum(a > 1 for a in data[channel_name].shape) == 1:
                    outs = wt.artists.quick1D(
                        data,
                        channel=channel_index,
                        autosave=True,
                        save_directory=output_path,
                        fname=image_fname,
                        verbose=False,
                    )
                else:
                    outs = wt.artists.quick2D(
                        data,
                        -1,
                        -2,
                        channel=channel_index,
                        autosave=True,
                        save_directory=output_path,
                        fname=image_fname,
                        verbose=False,
                    )
                if channel_name == main_channel:
                    outputs = outs
                data.transform(*orig_transform)
            # get output image
            if len(outputs) == 1:
                output_image_path = outputs[0]
            else:
                output_image_path = output_path / "animation.gif"
                wt.artists.stitch_to_animation(images=outputs, outpath=output_image_path)
            # upload
            self.upload(scan_folder, reference_image=str(output_image_path))

    def run(self):
        # axes
        axes = []
        for axis_name in self.aqn.read("scan", "axis names"):
            start = self.aqn.read(axis_name, "start")
            stop = self.aqn.read(axis_name, "stop")
            number = self.aqn.read(axis_name, "number")
            points = np.linspace(start, stop, int(number))
            units = self.aqn.read(axis_name, "units")
            axis = acquisition.Axis(points, units, axis_name)
            axes.append(axis)
        # constants
        constants = []
        for constant_name in self.aqn.read("scan", "constant names"):
            for hardware in all_hardwares:
                if hardware.name == constant_name:
                    units = hardware.units
                    if wt.units.is_valid_conversion(units, "wn"):
                        units = "wn"
                    if wt.units.is_valid_conversion(units, "ps"):
                        units = "ps"
                    break
            name = constant_name
            expression = self.aqn.read(constant_name, "expression")
            constant = acquisition.Constant(units, name, expression=expression, static=False)
            constants.append(constant)
        # do scan
        self.scan(axes, constants)
        # finish
        if not self.stopped.read():
            self.finished.write(True)  # only if acquisition successfull


### GUI #######################################################################


class GUI(acquisition.GUI):
    def add_axis(self, units):
        axis = Axis(units, len(self.axes))
        self.axes_container_widget.layout().addWidget(axis.widget)
        self.axes.append(axis)

    def add_buttons(self):
        colors = g.colors_dict.read()
        # layout
        button_container = QtWidgets.QWidget()
        button_container.setLayout(QtWidgets.QHBoxLayout())
        button_container.layout().setMargin(0)
        # remove
        remove_button = QtWidgets.QPushButton()
        remove_button.setText("REMOVE")
        remove_button.setMinimumHeight(25)
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors["stop"]
        )
        remove_button.setStyleSheet(StyleSheet)
        button_container.layout().addWidget(remove_button)
        # add
        add_button = QtWidgets.QPushButton()
        add_button.setText("ADD")
        add_button.setMinimumHeight(25)
        StyleSheet = "QPushButton{background:custom_color; border-width:0px;  border-radius: 0px; font: bold 14px}".replace(
            "custom_color", colors["set"]
        )
        add_button.setStyleSheet(StyleSheet)
        button_container.layout().addWidget(add_button)
        # finish
        self.layout.addWidget(button_container)
        return [add_button, remove_button]

    def add_constant(self):
        # if len(self.constants) == 1: return  # temporary...
        constant = Constant()
        self.constants_container_widget.layout().addWidget(constant.widget)
        self.constants.append(constant)

    def create_frame(self):
        # axes
        self.axes = []
        input_table = pw.InputTable()
        input_table.add("Axes", None)
        self.layout.addWidget(input_table)
        self.axes_container_widget = QtWidgets.QWidget()
        self.axes_container_widget.setLayout(QtWidgets.QVBoxLayout())
        self.axes_container_widget.layout().setMargin(0)
        self.layout.addWidget(self.axes_container_widget)
        add_energy_axis_button = pw.SetButton("ADD ENERGY AXIS")
        add_energy_axis_button.clicked.connect(lambda: self.add_axis("wn"))
        self.layout.addWidget(add_energy_axis_button)
        add_delay_axis_button = pw.SetButton("ADD DELAY AXIS")
        add_delay_axis_button.clicked.connect(lambda: self.add_axis("ps"))
        self.layout.addWidget(add_delay_axis_button)
        add_delay_axis_button = pw.SetButton("ADD ANGLE AXIS")
        add_delay_axis_button.clicked.connect(lambda: self.add_axis("deg"))
        self.layout.addWidget(add_delay_axis_button)
        remove_axis_button = pw.SetButton("REMOVE AXIS", "stop")
        remove_axis_button.clicked.connect(self.remove_axis)
        self.layout.addWidget(remove_axis_button)
        # constants
        self.constants = []
        input_table = pw.InputTable()
        input_table.add("Constants", None)
        self.layout.addWidget(input_table)
        self.constants_container_widget = QtWidgets.QWidget()
        self.constants_container_widget.setLayout(QtWidgets.QVBoxLayout())
        self.constants_container_widget.layout().setMargin(0)
        self.layout.addWidget(self.constants_container_widget)
        add_constant_button, remove_constant_button = self.add_buttons()
        add_constant_button.clicked.connect(self.add_constant)
        remove_constant_button.clicked.connect(self.remove_constant)
        # processing
        input_table = pw.InputTable()
        input_table.add("Processing", None)
        channel_names = list(yaqc_cmds.sensors.get_channels_dict().keys())
        if (
            "main_channel" not in self.state.keys()
            or self.state["main_channel"] not in channel_names
        ):
            self.state["main_channel"] = channel_names[0]
        self.channel_combo = pc.Combo(
            allowed_values=channel_names,
            initial_value=self.state["main_channel"],
        )
        self.channel_combo.updated.connect(self.save_state)
        input_table.add("Main Channel", self.channel_combo)
        if "process_all_channels" not in self.state.keys():
            self.state["process_all_channels"] = False
        self.process_all_channels = pc.Bool(initial_value=self.state["process_all_channels"])
        self.process_all_channels.updated.connect(self.save_state)
        input_table.add("Process All Channels", self.process_all_channels)
        self.layout.addWidget(input_table)

    def load(self, aqn_path):
        # clear old
        for axis in self.axes:
            axis.hide()
        for constant in self.constants:
            constant.hide()
        self.axes = []
        self.channels = []
        self.constants = []
        # read new
        aqn = wt.kit.INI(aqn_path)
        # axes
        axis_names = aqn.read("scan", "axis names")
        for axis_index, axis_name in enumerate(axis_names):
            units = aqn.read(axis_name, "units")
            if wt.units.is_valid_conversion(units, "wn"):
                units = "wn"
            elif wt.units.is_valid_conversion(units, "ps"):
                units = "ps"
            elif wt.units.is_valid_conversion(units, "deg"):
                units = "deg"
            axis = Axis(units, axis_index)
            axis.start.write(aqn.read(axis_name, "start"))
            axis.stop.write(aqn.read(axis_name, "stop"))
            axis.number.write(aqn.read(axis_name, "number"))
            hardwares = aqn.read(axis_name, "hardware")
            for hardware in hardwares:
                axis.hardwares[hardware].write(True)
            self.axes.append(axis)
            self.axes_container_widget.layout().addWidget(axis.widget)
        # constants
        constant_names = aqn.read("scan", "constant names")
        for constant_index, constant_name in enumerate(constant_names):
            constant = Constant()
            constant.hardware_name_combo.write(aqn.read(constant_name, "hardware"))
            constant.expression.write(aqn.read(constant_name, "expression"))
            self.constants.append(constant)
            self.constants_container_widget.layout().addWidget(constant.widget)
        # processing
        try:
            self.channel_combo.write(aqn.read("processing", "main channel"))
        except ValueError:
            pass  # TODO: log warning or something
        self.process_all_channels.write(aqn.read("processing", "process all channels"))
        # allow record to load settings
        # self.device_widget.load(aqn_path)

    def on_device_settings_updated(self):
        self.channel_combo.set_allowed_values(record.control.channel_names)

    def remove_axis(self):
        # remove trailing axis
        if len(self.axes) > 0:
            axis = self.axes[-1]
            self.axes_container_widget.layout().removeWidget(axis.widget)
            axis.hide()
            self.axes.pop(-1)

    def remove_constant(self):
        # remove trailing constant
        if len(self.constants) > 0:
            constant = self.constants[-1]
            self.constants_container_widget.layout().removeWidget(constant.widget)
            constant.hide()
            self.constants.pop(-1)

    def save(self, aqn_path):
        aqn = wt.kit.INI(aqn_path)
        # general
        axis_names = str([str(a.get_name()) for a in self.axes]).replace("'", "")
        aqn.write("info", "description", "SCAN: {}".format(axis_names))
        aqn.add_section("scan")
        aqn.write("scan", "axis names", [a.get_name() for a in self.axes])
        aqn.write("scan", "constant names", [c.get_name() for c in self.constants])
        # axes
        for axis in self.axes:
            name = axis.get_name()
            aqn.add_section(name)
            aqn.write(name, "start", axis.start.read())
            aqn.write(name, "stop", axis.stop.read())
            aqn.write(name, "number", axis.number.read())
            aqn.write(name, "units", axis.units)
            hardwares = []
            for key, bool_mutex in axis.hardwares.items():
                if bool_mutex.read():
                    hardwares.append(key)
            aqn.write(name, "hardware", hardwares)
        # constants
        for constant in self.constants:
            name = constant.get_name()
            aqn.add_section(name)
            aqn.write(name, "hardware", constant.hardware_name_combo.read())
            aqn.write(name, "expression", constant.expression.read())
        # processing
        aqn.add_section("processing")
        aqn.write("processing", "main channel", self.channel_combo.read())
        aqn.write("processing", "process all channels", self.process_all_channels.read())
        # allow devices to write settings
        # self.device_widget.save(aqn_path)

    def save_state(self):
        self.state["main_channel"] = self.channel_combo.read()
        self.state["process_all_channels"] = self.process_all_channels.read()
        super().save_state()


def load():
    return True


def mkGUI():
    global gui
    gui = GUI(module_name)



# Slacker implementation for yaqc_cmds


import os
import time
import tempfile
import pathlib

import appdirs
import toml

from PySide2 import QtGui, QtCore

import yaqc_cmds.project.classes as pc
import yaqc_cmds.project.logging_handler as logging_handler
import yaqc_cmds.project.project_globals as g
from . import bots


### container objects #########################################################


class Messages(QtCore.QMutex):
    def __init__(self):
        """
        holds list of enqueued options
        """
        QtCore.QMutex.__init__(self)
        self.value = []

    def read(self):
        self.lock()
        out = self.value
        self.value = []
        self.unlock()
        return out

    def push(self, value):
        self.lock()
        self.value.append(value)
        self.unlock()

    def pop(self):
        self.lock()
        self.value = self.value[1:]
        self.unlock()


messages_mutex = Messages()


### address & control objects #################################################


class Address(QtCore.QObject):
    update_ui = QtCore.Signal()
    queue_emptied = QtCore.Signal()

    def __init__(self, busy, enqueued):
        QtCore.QObject.__init__(self)
        self.busy = busy
        self.enqueued = enqueued
        self.ctrl = bots.witch
        self.name = "slack"

    def check_busy(self):
        """
        Handles writing of busy to False.

        Must always write to busy.
        """
        if self.is_busy():
            time.sleep(0.01)  # don't loop like crazy
            self.busy.write(True)
        elif self.enqueued.read():
            time.sleep(0.1)  # don't loop like crazy
            self.busy.write(True)
        else:
            self.busy.write(False)
            self.update_ui.emit()

    @QtCore.Slot(str, list)
    def dequeue(self, method, inputs):
        """
        Slot to accept enqueued commands from main thread.

        Method passed as qstring, inputs as list of [args, kwargs].

        Calls own method with arguments from inputs.
        """
        self.update_ui.emit()
        method = str(method)  # method passed as qstring
        args, kwargs = inputs
        self.enqueued.pop()
        getattr(self, method)(*args, **kwargs)
        if not self.enqueued.read():
            self.queue_emptied.emit()
            self.check_busy()

    def get_messages(self, inputs):
        newer_than = inputs[0]
        messages = self.ctrl.get_messages(newer_than=newer_than)
        for message in messages:
            messages_mutex.push(message)

    def is_busy(self):
        return False

    def send_message(self, inputs):
        text, channel, attachments = inputs
        self.ctrl.send_message(text, channel, attachments)

    def upload_file(self, inputs):
        file_path, Title, first_comment, channel = inputs
        self.ctrl.upload_file(file_path, Title, first_comment, channel)

    def delete_files(self, inputs):
        self.ctrl.delete_files

    def sign_off(self, inputs):
        message = inputs[0]
        self.ctrl.send_message(message)


class Control:
    def __init__(self):
        self.most_recent_delete = time.time()
        self.config = toml.load(
            pathlib.Path(appdirs.user_config_dir("yaqc-cmds", "yaqc-cmds")) / "config.toml"
        )
        # create control containers
        self.busy = pc.Busy()
        self.enqueued = pc.Enqueued()
        # create address object
        self.thread = QtCore.QThread()
        self.address = Address(self.busy, self.enqueued)
        self.address.moveToThread(self.thread)
        self.thread.start()
        # create q
        self.q = pc.Q(self.enqueued, self.busy, self.address)
        # connect
        g.shutdown.add_method(self.close)
        # signal startup
        self.send_message(":wave: signing on", self.config["slack"]["channel"])
        g.slack_control.write(self)

    def append(self, text, channel):
        self.send_message(":confounded: sorry, that feature hasn't been implemented")

    def close(self):
        self.q.push("sign_off", [":spock-hand: signing off"])
        time.sleep(1)
        # quit thread
        self.thread.exit()
        self.thread.quit()

    def get(self, text, channel):
        # interpret text as integer
        try:
            i = int(text)
        except:
            self.send_message(":interrobang: I couldn't find a number in your request")
            return

    def interrupt(self, text, channel):
        subcommand = text.strip().upper()
        if subcommand == "":
            subcommand = "PAUSE"
        messages = {
            "PAUSE": ":double_vertical_bar: Queue Paused, use `interrupt resume` to continue",
            "RESUME": ":arrow_forward: Queue resumed",
            "STOP": ":octagonal_sign: Queue stopped, use `run` to continue with next item",
            "SKIP": ":black_right_pointing_double_triangle_with_vertical_bar: Item skipped, continuing queue",
        }
        message = messages.get(subcommand, ":confounded: I do not understand the command")
        if subcommand == "PAUSE":
            g.main_window.read().queue_gui.queue.status.pause.write(True)
        elif subcommand in messages.keys():
            if g.main_window.read().queue_gui.queue.status.going.read():
                g.main_window.read().queue_gui.queue.interrupt(option=subcommand)
            elif subcommand == "STOP":
                g.main_window.read().queue_gui.queue.status.go.write(False)
        self.send_message(message)
        g.main_window.read().queue_gui.update_ui()

    def run_queue(self):
        if not g.main_window.read().queue_gui.queue.status.go.read():
            g.main_window.read().queue_gui.queue.status.go.write(True)
            try:
                g.main_window.read().queue_gui.queue.run()
            except IndexError:  # Queue full
                pass
        g.main_window.read().queue_gui.update_ui()

    def screenshot(self, channel):
        p = QtGui.QPixmap.grabWindow(g.main_window.read().winId())
        tf = tempfile.mkstemp()
        p.save(tf[1], "png")
        self.upload_file(tf[1], ":camera:", channel=channel)
        time.sleep(1)
        os.unlink(tf[1])

    def log(self, text, channel):
        log_filepath = logging_handler.filepath
        if "get" in text:
            self.upload_file(log_filepath, channel=channel)
            return
        # extract numbers from text
        numbers = [int(s) for s in text.split() if s.isdigit()]
        if len(numbers) > 0:
            number = numbers[0]
        else:
            number = 10
        # get log text
        with open(log_filepath) as f:
            lines = f.readlines()
        lines.reverse()
        num_lines = len(lines)
        if len(lines) < number:
            number = len(lines)
        lines = lines[:number]
        # send message
        list_string = self._make_list(lines)
        attachment = self._make_attachment(list_string)
        self.send_message(
            "here are the {0} most recent log items (out of {1})".format(number, num_lines),
            channel,
            [attachment],
        )

    def ls(self, text, channel):
        # extract numbers from text
        numbers = [int(s) for s in text.split() if s.isdigit()]
        if len(numbers) > 0:
            number = numbers[0]
        else:
            number = 10
        # get files
        folder_names = self._get_data_folders()
        if len(folder_names) == 0:
            self.send_message("yaqc_cmds currently has no data")
            return
        num_folders = len(folder_names)
        if len(folder_names) < number:
            number = len(folder_names)
        folder_names = folder_names[:number]
        # send message
        list_string = self._make_list(folder_names)
        attachment = self._make_attachment(list_string)
        self.send_message(
            "here are the {0} most recent aquisitions (out of {1})".format(number, num_folders),
            channel,
            [attachment],
        )

    def make_attachment(self, text, title=None, pretext=None, fields=None, color="#808080"):
        attachment = {}
        if pretext is not None:
            attachment["pretext"] = pretext
        if fields is not None:
            attachment["fields"] = fields
        if title is not None:
            attachment["title"] = title
        attachment["color"] = color
        attachment["text"] = text
        attachment["mrkdwn_in"] = ["fields"]
        return attachment

    def make_field(self, title, value, short=False):
        field = {}
        field["title"] = title
        field["value"] = value
        field["short"] = short
        return field

    def move(self, text, channel):
        self.send_message(":confounded: sorry, that feature hasn't been implemented")

    def poll(self):
        if not self.busy.read():
            self.q.push("get_messages", [60])
        self.read_messages()

    def read_messages(self):
        messages = messages_mutex.read()
        for message in messages:
            # only process messages
            if not "type" in message.keys():
                continue
            if not message["type"] == "message":
                continue
            # unpack some things
            text = message["text"]
            channel = message["channel"]
            # only process messages that are posted in the appropriate channel
            if not channel == self.config["slack"]["channel"]:
                continue
            text = text.split(" ", 1)[1]
            print("message read from slack:", text)
            # process
            if "echo " in text.lower():
                out = text.split("echo ", 1)[1]
                message = ":mega: " + out
                self.send_message(message, channel)
            elif "get" == text[:3].lower():
                self.get(text, channel)
            elif "status" in text.lower():
                self.status(text, channel)
            elif "remove" in text.lower():
                self.remove(text, channel)
            elif "move" in text.lower():
                self.move(text, channel)
            elif "append" in text.lower():
                self.append(text, channel)
            elif "interrupt" in text.lower():
                try:
                    text = text.split(" ", 1)[1]
                except IndexError:
                    text = ""
                self.interrupt(text, channel)
            elif "run" in text.lower():
                self.run_queue()
            elif "screenshot" in text.lower():
                self.screenshot(channel)
            elif "help" in text.lower():
                self.send_help(channel)
            else:
                self.send_message(
                    ":thinking_face: command '{}' not recognized - type 'help' for a list of available commands".format(
                        text
                    ),
                    channel,
                )

    def remove(self, text, channel):
        self.send_message(":confounded: sorry, that feature hasn't been implemented")

    def send_help(self, channel):
        command_fields = []
        command_fields.append(
            self.make_field("status [--full]", "Get the current status of yaqc_cmds.")
        )
        # command_fields.append(self.make_field('remove i', 'Remove the ith item from the queue.'))
        # command_fields.append(self.make_field('move i to j', 'Move item i to position j. All other items retain their order.'))
        # command_fields.append(self.make_field('append [name] [info]', 'Append a file to the queue. Must be made as a comment of an attached file.'))
        command_fields.append(self.make_field("run", "Run the queue."))
        command_fields.append(
            self.make_field(
                "interrupt [pause|resume|skip|stop]",
                "Interrupt the queue or resume from pause. Default is pause.",
            )
        )
        command_fields.append(
            self.make_field(
                "screenshot",
                "Take a screenshot of the current window and post to slack.",
            )
        )
        command_fields.append(self.make_field("help", "Show this help message."))
        attachment = self.make_attachment("", fields=command_fields)
        self.send_message(":robot_face: here are my commands", channel, [attachment])

    def send_message(self, text, channel=None, attachments=[]):
        self.q.push("send_message", [text, channel, attachments])

    def status(self, text, channel):
        text, attachments = g.main_window.read().get_status("full" in text.lower())
        self.send_message(text, channel, attachments)

    def upload_file(self, file_path, title=None, first_comment=None, channel=None):
        self.q.push("upload_file", [file_path, title, first_comment, channel])


control = Control()



import attune
import WrightTools as wt

import somatic.modules.abstract_tuning as abstract_tuning


module_name = "TUNE INTENSITY"


class Worker(abstract_tuning.Worker):
    reference_image = "intensity.png"

    def _process(self, data, curve, channel, gtol, ltol, level, scan_folder, config):
        opa = config["OPA"]["opa"]
        dep = config["Motor"]["motor"]
        data[channel].signed = False
        transform = list(data.axis_expressions)
        transform = transform[:2]
        if level:
            data.level(channel, -1, 5)
        for axis in data.axis_expressions[2:]:
            data.moment(
                axis,
                channel,
                moment=0,
                resultant=wt.kit.joint_shape(data[opa], data[f"{opa}_{dep}_points"]),
            )
            channel = -1
        data.transform(opa, f"{opa}_{dep}_points")
        return attune.workup.intensity(
            data,
            channel,
            dep,
            curve,
            level=level,
            gtol=gtol,
            ltol=ltol,
            save_directory=scan_folder,
        )


class GUI(abstract_tuning.GUI):
    def __init__(self, module_name):
        self.items = {}
        self.items["Motor"] = abstract_tuning.MotorAxisSectionWidget("Motor", self)
        super().__init__(module_name)
        self.items["Spectrometer"]["Action"].set_allowed_values(
            self.items["Spectrometer"]["Action"].allowed_values[1:]
        )


def load():
    return True


def mkGUI():
    global gui
    gui = GUI(module_name)



from qtpy import QtCore


class SignalContainer(QtCore.QObject):
    data_file_created = QtCore.Signal()
    data_file_written = QtCore.Signal()
    queue_relinquishing_control = QtCore.Signal()
    queue_taking_control = QtCore.Signal()
    updated_attune_store = QtCore.Signal()


_signal_container = SignalContainer()
data_file_created = _signal_container.data_file_created
data_file_written = _signal_container.data_file_written
queue_relinquishing_control = _signal_container.queue_relinquishing_control
queue_taking_control = _signal_container.queue_taking_control
updated_attune_store = _signal_container.updated_attune_store



"""Sensors."""


import appdirs
import pathlib
import time

import collections

import numpy as np
import toml

from PySide2 import QtCore, QtWidgets

import WrightTools as wt
import yaqc

import yaqc_cmds
import yaqc_cmds.project.classes as pc
import yaqc_cmds.project.widgets as pw


class Data(QtCore.QMutex):
    def __init__(self):
        QtCore.QMutex.__init__(self)
        self.WaitCondition = QtCore.QWaitCondition()
        self.channels = {}
        self.signed = []

    def read(self):
        return self.channels

    def write(self, channels):
        self.lock()
        self.channels = channels
        self.WaitCondition.wakeAll()
        self.unlock()

    def write_properties(self, channels, signed=False):
        self.lock()
        self.channels = channels
        self.signed = signed
        if not signed:
            self.signed = [False] * len(self.channels)
        self.WaitCondition.wakeAll()
        self.unlock()

    def wait_for_update(self, timeout=5000):
        if self.value:
            self.lock()
            self.WaitCondition.wait(self, timeout)
            self.unlock()


class Sensor(pc.Hardware):
    settings_updated = QtCore.Signal()

    def __init__(self, *args, **kwargs):
        self.freerun = pc.Bool(initial_value=False)
        self.Widget = kwargs.pop("Widget")
        self.data = Data()
        self.active = True
        self.measure_time = pc.Number(initial_value=np.nan, display=True, decimals=3)
        super().__init__(*args, **kwargs)
        self.settings_updated.emit()
        self.freerun.write(True)
        self.on_freerun_updated()

    @property
    def channel_names(self):
        return list(self.data.channels.keys())

    @property
    def channels(self):
        return self.data.channels

    def get_headers(self):
        out = collections.OrderedDict()
        return out

    def give_widget(self, widget):
        self.widget = widget
        self.gui.create_frame(widget)

    def initialize(self):
        # self.wait_until_still()
        self.freerun.updated.connect(self.on_freerun_updated)
        self.update_ui.emit()
        self.driver.update_ui.connect(self.on_driver_update_ui)
        self.settings_updated.emit()

    def load_settings(self, aqn):
        pass

    def measure(self):
        self.q.push("measure")

    def on_driver_update_ui(self):
        self.update_ui.emit()

    def on_freerun_updated(self):
        self.q.push("loop")

    def set_freerun(self, state):
        self.freerun.write(state)
        self.on_freerun_updated()
        self.settings_updated.emit()  # TODO: should probably remove this

    def wait_until_still(self):
        while self.busy.read():
            print(f"{self.name} is busy")
            self.busy.wait_for_update()


class Driver(pc.Driver):
    settings_updated = QtCore.Signal()

    def __init__(self, sensor, yaqd_port):
        super().__init__()
        self.client = yaqc.Client(yaqd_port)
        # attributes
        self.name = self.client.id()["name"]
        self.enqueued = sensor.enqueued
        self.busy = sensor.busy
        self.freerun = sensor.freerun
        self.data = sensor.data
        self.measure_time = sensor.measure_time
        self.thread = sensor.thread

    def initialize(self):
        self.measure()
        yaqc_cmds.sensors.signals.sensors_changed.emit()
        yaqc_cmds.sensors.signals.channels_changed.emit()

    def loop(self):
        while self.freerun.read() and not self.enqueued.read():
            initial_time = time.time()
            self.measure()
            self.busy.write(False)
            # Rate limit when just freerunning
            time.sleep(max(0.1 - time.time() + initial_time, 0))

    def measure(self):
        timer = wt.kit.Timer(verbose=False)
        with timer:
            self.busy.write(True)
            if "has-measure-trigger" in self.client.traits:
                self.client.measure(loop=False)
                while self.client.busy():
                    time.sleep(0.01)
            out = self.client.get_measured()
            del out["measurement_id"]
            try:
                del out["mapping_id"]
            except KeyError:
                # No mapping
                pass
            signed = [False for _ in out]
            self.data.write_properties(out, signed)
            self.busy.write(False)
        self.measure_time.write(timer.interval)
        self.update_ui.emit()

    def shutdown(self):
        pass


class SensorWidget(QtWidgets.QWidget):
    def __init__(self):
        QtWidgets.QWidget.__init__(self)

    def load(self, aqn_path):
        # TODO:
        pass

    def save(self, aqn_path):
        # TODO:
        ini = wt.kit.INI(aqn_path)
        ini.add_section("Virtual")
        ini.write("Virtual", "use", True)


class Widget(QtWidgets.QWidget):
    def __init__(self):
        QtWidgets.QWidget.__init__(self)
        layout = QtWidgets.QVBoxLayout()
        self.setLayout(layout)
        layout.setMargin(0)
        input_table = pw.InputTable()
        input_table.add("Virtual", None)
        self.use = pc.Bool(initial_value=True)
        input_table.add("Use", self.use)
        layout.addWidget(input_table)

    def load(self, aqn_path):
        pass

    def save(self, aqn_path):
        ini = wt.kit.INI(aqn_path)
        ini.add_section("virtual")
        ini.write("virtual", "use", self.use.read())



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['yaqc_cmds',
 'yaqc_cmds.autonomic',
 'yaqc_cmds.hardware',
 'yaqc_cmds.project',
 'yaqc_cmds.project.slack',
 'yaqc_cmds.sensors',
 'yaqc_cmds.somatic',
 'yaqc_cmds.somatic.order']

package_data = \
{'': ['*'],
 'yaqc_cmds.project': ['widget files/*'],
 'yaqc_cmds.somatic': ['modules/*']}

install_requires = \
['appdirs',
 'attune',
 'click',
 'numexpr',
 'pyqtgraph',
 'pyside2',
 'pyserial',
 'qtpy',
 'slackclient<2',
 'slacker',
 'sympy',
 'toml',
 'wrighttools',
 'yaqc']

entry_points = \
{'console_scripts': ['yaqc-cmds = yaqc_cmds.__main__:main']}

setup(name='yaqc-cmds',
      version='2021.3.0',
      description='Unified software for controlling hardware and collecting data in the Wright group.',
      author='Blaise Thompson',
      author_email='blaise@untzag.com',
      url='https://github.com/wright-group/yaqc-cmds',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.7',
     )



"""wt5 data file functions"""


import time
import threading

import h5py
import WrightTools as wt

import yaqc_cmds.project.project_globals as g
import yaqc_cmds.somatic as somatic


class DataContainer(object):
    def __init__(self):
        self._data = None
        self.data_filepath = None
        self.last_idx_written = None
        self.lock = threading.RLock()

    def __enter__(self):
        self.lock.acquire()
        if self.data_filepath:
            self._data = wt.open(self.data_filepath, edit_local=True)
        return self._data

    def __exit__(self, exc_type, exc_value, traceback):
        if self._data:
            self._data.close()
        self._data = None
        self.lock.release()


data_container = DataContainer()


def create_data(path, headers, destinations, axes, constants, hardware, sensors):
    """Create new data object.

    Parameters
    ----------
    path : path-like
        Full path to new file.
    headers : dictionary
        Metadata
    destinations : list of yaqc_cmds.acquisition.Destination objects
        New scan destinations.
    axes : list of yaqc_cmds.acqusition.Axis objects
        New scan axes.
    constants : list of yaqc_cmds.acquisition.Constant objects
        New scan constants.
    hardware: list of yaqc_cmds.hardware.Hardware objects
        all active hardware
    sensors: list of yaqc_cmds._sensors.Sensor objects
        all active sensors
    """
    f = h5py.File(path, "w")
    global data_container
    with data_container as data:
        data = wt.Data(f, name=headers["name"], edit_local=True)
        data_container.data_filepath = path

        # fill out yaqc_cmds_information in headers
        headers["Yaqc_cmds version"] = g.version.read()
        headers["system name"] = g.system_name.read()

        data.attrs.update(headers)

        full_scan_shape = tuple(a.points.size for a in axes)
        variable_shapes = {"labtime": full_scan_shape}
        variable_units = {"labtime": "s"}
        variable_labels = {}

        for i, axis in enumerate(axes):
            shape = [1] * len(axes)
            shape[i] = axis.points.size
            variable_shapes[f"{axis.name}_points"] = tuple(shape)
            variable_units[f"{axis.name}_points"] = axis.units
            if hasattr(axis, "centers"):
                shape = list(full_scan_shape)
                shape[i] = 1
                variable_shapes[f"{axis.name}_centers"] = tuple(shape)
                variable_units[f"{axis.name}_centers"] = axis.units

        for hw in hardware:
            for rec, (_, units, _, label, _) in hw.recorded.items():
                variable_shapes[rec] = full_scan_shape
                variable_units[rec] = units
                variable_labels[rec] = label

        channel_shapes = {}
        channel_units = {}

        transform_extras = []

        for sensor in sensors:
            # TODO allow sensors to be inactive
            sensor.active = True

            if "has-mapping" in sensor.driver.client.traits:
                channel_mappings = sensor.driver.client.get_channel_mappings()
                shapes = sensor.driver.client.get_channel_shapes()
                chs_left = set(channel_mappings.keys())

                variable_units.update(sensor.driver.client.get_mapping_units())
                channel_units.update(sensor.driver.client.get_channel_units())
                mappings = sensor.driver.client.get_mappings()

                while chs_left:
                    chs_broad = [chs_left.pop()]
                    maps_broad = channel_mappings[chs_broad[0]]
                    ndim = len(shapes[chs_broad[0]])
                    for ch in chs_left:
                        if any(i in mappings[ch] for i in maps_broad):
                            maps_broad.extend(channel_mappings[ch])
                            chs_broad.append(ch)
                            chs_left.remove(ch)
                    for k, v in variable_shapes.items():
                        variable_shapes[k] = v + (1,) * ndim
                    for k, v in channel_shapes.items():
                        channel_shapes[k] = v + (1,) * ndim
                    variable_shapes.update(
                        {m: full_scan_shape + mappings[m].shape for m in maps_broad}
                    )
                    transform_extras.extend(maps_broad)
                    channel_shapes.update(
                        {ch: full_scan_shape + tuple(shapes[ch]) for ch in chs_broad}
                    )
                    full_scan_shape += (1,) * ndim
                # TODO channels with _no_ mapping?

            else:
                for ch in sensor.channel_names:
                    channel_shapes[ch] = full_scan_shape
                    # TODO: channel units?
                    channel_units[ch] = None

        for var, sh in variable_shapes.items():
            units = variable_units[var]
            label = variable_labels.get(var)
            if label:
                data.create_variable(var, shape=sh, units=units, label=label)
            else:
                data.create_variable(var, shape=sh, units=units)

        for axis in axes:
            sh = data[f"{axis.name}_points"].shape
            data[f"{axis.name}_points"][:] = axis.points.reshape(sh)
            if hasattr(axis, "centers"):
                sh = data[f"{axis.name}_centers"].shape
                data[f"{axis.name}_centers"][:] = axis.centers.reshape(sh)

        # This check was originally if there was _centers_ use the points arrays
        # This was changed to always use the points arrays (except for the array detector)
        # Because chopping full shaped arrays caused incorrect behavior for 3D+ data
        # When we have some better hinting in WT itself, this should likely be reverted
        # KFS 2020-11-13
        transform = [f"{a.name}_points" if hasattr(a, "points") else a.name for a in axes]
        transform.extend(transform_extras)
        data.transform(*transform)

        for ch, sh in channel_shapes.items():
            units = channel_units[ch]
            data.create_channel(ch, shape=sh, units=units)
            # TODO signed?
            # TODO labels?

        somatic.signals.data_file_created.emit()


def write_data(idx, hardware, sensors):
    global data_container
    with data_container as data:
        in_idx = idx
        idx = idx + (...,)
        data["labtime"][idx] = time.time()
        for hw in hardware:
            for rec, (obj, *_) in hw.recorded.items():
                data[rec][idx] = obj.read(data[rec].units)
        for s in sensors:
            if "has-mapping" in s.driver.client.traits:
                for ch, val in s.channels.items():
                    idx_ch = list(data[ch].shape)
                    print(idx_ch)
                    idx_ch = [slice(None) if i != 1 else 1 for i in idx_ch]
                    idx_ch[: len(in_idx)] = in_idx
                    idx_ch = tuple(idx_ch)
                    print(data[ch].shape, idx_ch, val.shape)
                    data[ch][idx_ch] = val
                for var, val in s.driver.client.get_mappings().items():
                    if var == "mapping_id":
                        continue
                    idx_map = list(data[var].shape)
                    idx_map = [slice(None) if i != 1 else 1 for i in idx_map]
                    idx_map[: len(in_idx)] = in_idx
                    idx_map = tuple(idx_map)
                    print(data[var].shape, idx_map, val.shape)
                    data[var][idx_map] = val
            else:
                for ch, val in s.channels.items():
                    data[ch][idx] = val
        data.flush()
        data_container.last_idx_written = in_idx



"""
a collection of small, general purpose objects and methods
"""


### import ####################################################################


import numpy as np


### math ######################################################################


def grating_linear_dispersion(
    spec_inclusion_angle,
    spec_focal_length,
    spec_focal_length_tilt,
    spec_grooves_per_mm,
    spec_central_wavelength,
    spec_order,
    number_of_pixels,
    pixel_width,
    calibration_pixel,
):
    """
    Parameters
    ----------
    spec_inclusion_angle : float
        Spectrometer inclusion angle (degrees).
    spec_focal_length : float
        Spectrometer focal length (mm).
    spec_focal_length_tilt : float
        Spectrometer focal length tilt (degrees).
    spec_grooves_per_mm : float
        Spectrometer grating grooves per mm.
    spec_central_wavelength : float
        Spectrometer set wavelength (nm).
    spec_order : int
        Spectrometer order of diffraction.
    number_of_pixels : int
        Number of pixels on array detector.
    pixel_width : float
        Pixel width (um)
    calibration_pixel : int
        Pixel on which the spec_central_wavelength falls.

    Returns
    -------
    np.ndarray
        Color found on each pixel (nm)
    """
    # translate inputs into appropriate internal units
    spec_inclusion_angle_rad = np.radians(spec_inclusion_angle)
    spec_focal_length_tilt_rad = np.radians(spec_focal_length_tilt)
    pixel_width_mm = pixel_width / 1e3
    # create array
    i_pixel = np.arange(number_of_pixels)
    # calculate terms
    x = np.arcsin(
        (1e-6 * spec_order * spec_grooves_per_mm * spec_central_wavelength)
        / (2 * np.cos(spec_inclusion_angle_rad / 2.0))
    )
    A = np.sin(x - spec_inclusion_angle_rad / 2)
    B = np.sin(
        (spec_inclusion_angle_rad)
        + x
        - (spec_inclusion_angle_rad / 2)
        - np.arctan(
            (
                pixel_width_mm * (i_pixel - calibration_pixel)
                + spec_focal_length * spec_focal_length_tilt_rad
            )
            / (spec_focal_length * np.cos(spec_focal_length_tilt_rad))
        )
    )
    out = ((A + B) * 1e6) / (spec_order * spec_grooves_per_mm)
    return out


### testing ###################################################################

if __name__ == "__main__":
    arr = grating_linear_dispersion(
        spec_inclusion_angle=24.0,
        spec_focal_length=140.0,
        spec_focal_length_tilt=0.0,
        spec_grooves_per_mm=300.0,
        spec_central_wavelength=1300.0,
        spec_order=1,
        number_of_pixels=256,
        pixel_width=50.0,
        calibration_pixel=100,
    )
    print(arr[100])



import attune

import somatic.modules.abstract_tuning as abstract_tuning

module_name = "TUNE TEST"


class Worker(abstract_tuning.Worker):
    reference_image = "tune_test.png"

    def _process(self, data, curve, channel, gtol, ltol, level, scan_folder, config):
        opa = config["OPA"]["opa"]
        spec = config["Spectral Axis"]["axis"]
        data.transform(opa, f"{spec}-{opa}")
        return attune.workup.tune_test(
            data,
            channel,
            curve,
            level=level,
            gtol=gtol,
            ltol=ltol,
            save_directory=scan_folder,
        )


class GUI(abstract_tuning.GUI):
    def __init__(self, module_name):
        self.items = {}
        self.items["Spectral Axis"] = abstract_tuning.SpectralAxisSectionWidget(
            "Spectral Axis", self
        )
        super().__init__(module_name)


def load():
    return True


def mkGUI():
    global gui
    gui = GUI(module_name)



### import ####################################################################


import pathlib

import numpy as np

import matplotlib

matplotlib.pyplot.ioff()

import WrightTools as wt

import yaqc_cmds.project.classes as pc
import yaqc_cmds.project.widgets as pw
import yaqc_cmds.somatic.acquisition as acquisition

import yaqc_cmds
import yaqc_cmds.hardware.opas as opas
import yaqc_cmds.hardware.spectrometers as spectrometers
import yaqc_cmds.sensors as sensors
from yaqc_cmds.somatic import _wt5


### define ####################################################################


module_name = "MOTORTUNE"


### custom classes ############################################################


class MotorGUI:
    def __init__(self, name, center, width, number, use_tune_points):
        self.name = name
        self.use_tune_points = use_tune_points
        self.input_table = pw.InputTable()
        self.input_table.add(name, None)
        allowed = ["Set", "Scan", "Static"]
        self.method = pc.Combo(allowed_values=allowed)
        if self.use_tune_points is not None:
            self.use_tune_points.updated.connect(self.update_disabled)
        self.method.updated.connect(self.update_disabled)
        self.input_table.add("Method", self.method)
        self.center = pc.Number(initial_value=center)
        self.input_table.add("Center", self.center)
        self.width = pc.Number(initial_value=width)
        self.input_table.add("Width", self.width)
        self.npts = pc.Number(initial_value=number, decimals=0)
        self.input_table.add("Number", self.npts)
        self.update_disabled()

    def update_disabled(self):
        self.center.set_disabled(True)
        self.width.set_disabled(True)
        self.npts.set_disabled(True)
        method = self.method.read()
        if method == "Set":
            self.center.set_disabled(self.use_tune_points.read())
        elif method == "Scan":
            self.center.set_disabled(self.use_tune_points.read())
            self.width.set_disabled(False)
            self.npts.set_disabled(False)
        elif method == "Static":
            self.center.set_disabled(False)


class OPA_GUI:
    def __init__(self, hardware, layout, use_tune_points):
        self.hardware = hardware
        motor_names = self.hardware.motor_names
        self.motors = []
        for name in motor_names:
            motor = MotorGUI(name, 30, 1, 11, use_tune_points)
            if layout is not None:
                layout.addWidget(motor.input_table)
            self.motors.append(motor)
        self.hide()  # initialize hidden

    def hide(self):
        for motor in self.motors:
            motor.input_table.hide()

    def show(self):
        for motor in self.motors:
            motor.input_table.show()


### Worker ####################################################################


class Worker(acquisition.Worker):
    def process(self, scan_folder):
        with _wt5.data_container as data:
            # decide which channels to make plots for
            channel_name = self.aqn.read("processing", "channel")
            # make figures for each channel
            data_path = pathlib.Path(_wt5.data_container.data_filepath)
            data_folder = data_path.parent
            # make all images
            channel_path = data_folder / channel_name
            output_path = data_folder
            if data.ndim > 2:
                output_path = channel_path
                channel_path.mkdir()
            image_fname = channel_name
            if data.ndim == 1:
                outs = wt.artists.quick1D(
                    data,
                    channel=channel_name,
                    autosave=True,
                    save_directory=output_path,
                    fname=image_fname,
                    verbose=False,
                )
            else:
                outs = wt.artists.quick2D(
                    data,
                    -1,
                    -2,
                    channel=channel_name,
                    autosave=True,
                    save_directory=output_path,
                    fname=image_fname,
                    verbose=False,
                )
        # get output image
        if len(outs) == 1:
            output_image_path = outs[0]
        else:
            output_image_path = output_path / "animation.gif"
            wt.artists.stitch_to_animation(images=outs, outpath=output_image_path)
        # upload
        self.upload(scan_folder, reference_image=str(output_image_path))

    def run(self):
        # assemble axes
        axes = []
        # get OPA properties
        opa_name = self.aqn.read("motortune", "opa name")
        opa_names = [h.name for h in opas.hardwares]
        opa_index = opa_names.index(opa_name)
        opa_hardware = opas.hardwares[opa_index]
        opa_friendly_name = opa_hardware.name
        curve = opa_hardware.curve
        arrangement = opa_hardware.curve.arrangements[opa_hardware.arrangement]
        motor_names = self.aqn.read("motortune", "motor names")
        scanned_motors = [m for m in motor_names if self.aqn.read(m, "method") == "Scan"]
        tune_points = get_tune_points(curve, arrangement, scanned_motors)
        tune_units = "nm"  # needs update if/when attune supports other units for independents
        # tune points
        if self.aqn.read("motortune", "use tune points"):
            motors_excepted = []  # list of indicies
            for motor_name in motor_names:
                if not self.aqn.read(motor_name, "method") == "Set":
                    motors_excepted.append(motor_name)
            if self.aqn.read("spectrometer", "method") == "Set":
                hardware_dict = {
                    opa_friendly_name: [
                        opa_hardware,
                        "set_position_except",
                        ["destination", motors_excepted, "units"],
                    ],
                    "wm": [spectrometers.hardwares[0], "set_position", None],
                }
                axis = acquisition.Axis(
                    tune_points,
                    tune_units,
                    opa_friendly_name,
                    hardware_dict,
                )
                axes.append(axis)
            else:
                hardware_dict = {
                    opa_friendly_name: [
                        opa_hardware,
                        "set_position_except",
                        ["destination", motors_excepted, "units"],
                    ]
                }
                axis = acquisition.Axis(
                    tune_points,
                    tune_units,
                    opa_friendly_name,
                    hardware_dict,
                )
                axes.append(axis)
        # motor
        for motor_name in motor_names:
            if self.aqn.read(motor_name, "method") == "Scan":
                motor_units = None
                name = "_".join([opa_friendly_name, motor_name])
                width = self.aqn.read(motor_name, "width") / 2.0
                npts = int(self.aqn.read(motor_name, "number"))
                if self.aqn.read("motortune", "use tune points"):
                    center = 0.0
                    kwargs = {
                        "centers": [curve(t, arrangement.name)[motor_name] for t in tune_points]
                    }
                else:
                    center = self.aqn.read(motor_name, "center")
                    kwargs = {}
                points = np.linspace(center - width, center + width, npts)
                hardware_dict = {name: [opa_hardware, "set_motor", [motor_name, "destination"]]}
                axis = acquisition.Axis(points, motor_units, name, hardware_dict, **kwargs)
                axes.append(axis)
            elif self.aqn.read(motor_name, "method") == "Set":
                pass
            elif self.aqn.read(motor_name, "method") == "Static":
                opa_hardware.q.push("set_motor", [motor_name, self.aqn.read(motor_name, "center")])
        # mono
        if self.aqn.read("spectrometer", "method") == "Scan":
            name = "wm"
            units = "wn"
            width = self.aqn.read("spectrometer", "width") / 2.0
            npts = int(self.aqn.read("spectrometer", "number"))
            if self.aqn.read("motortune", "use tune points"):
                center = 0.0
                kwargs = {"centers": wt.units.convert(tune_points, tune_units, units)}
            else:
                center = self.aqn.read("spectrometer", "center")
                center = wt.units.convert(
                    center, self.aqn.read("spectrometer", "center units"), "wn"
                )
                kwargs = {}
            points = np.linspace(center - width, center + width, npts)
            axis = acquisition.Axis(points, units, name, **kwargs)
            axes.append(axis)
        elif self.aqn.read("spectrometer", "method") == "Set":
            if self.aqn.read("motortune", "use tune points"):
                # already handled above
                pass
            else:
                center = self.aqn.read("spectrometer", "center")
                center = wt.units.convert(
                    center, self.aqn.read("spectrometer", "center units"), "wn"
                )
                spectrometers.hardwares[0].set_position(center, "wn")
        elif self.aqn.read("spectrometer", "method") == "Static":
            center = self.aqn.read("spectrometer", "center")
            center = wt.units.convert(center, self.aqn.read("spectrometer", "center units"), "wn")
            spectrometers.hardwares[0].set_position(center, "wn")
        # handle centers
        for axis_index, axis in enumerate(axes):
            centers_shape = [a.points.size for i, a in enumerate(axes) if not i == axis_index]
            ones = np.ones(centers_shape)
            if hasattr(axis, "centers"):
                # arrays always follow
                axis.centers = np.transpose(axis.centers * ones.T)
        # launch
        pre_wait_methods = [
            lambda: opa_hardware.q.push("wait_until_still"),
            lambda: opa_hardware.q.push("get_motor_positions"),
            lambda: opa_hardware.q.push("get_position"),
        ]
        # do scan
        self.scan(axes, constants=[], pre_wait_methods=pre_wait_methods)
        if not self.stopped.read():
            self.finished.write(True)  # only if acquisition successfull


### GUI #######################################################################


class GUI(acquisition.GUI):
    def create_frame(self):
        # shared settings
        input_table = pw.InputTable()
        allowed = [hardware.name for hardware in opas.hardwares]
        self.opa_combo = pc.Combo(allowed)
        input_table.add("OPA", self.opa_combo)
        self.opa_combo.updated.connect(self.on_opa_combo_updated)
        self.use_tune_points = pc.Bool(initial_value=True)
        input_table.add("Use Tune Points", self.use_tune_points)
        self.layout.addWidget(input_table)
        # motor settings
        self.opa_guis = [
            OPA_GUI(hardware, self.layout, self.use_tune_points) for hardware in opas.hardwares
        ]
        self.opa_guis[0].show()
        # mono settings
        allowed = ["Set", "Scan", "Static"]
        self.mono_method_combo = pc.Combo(allowed, disable_under_module_control=True)
        self.mono_method_combo.updated.connect(self.update_mono_settings)
        self.mono_center = pc.Number(
            initial_value=7000, units="wn", disable_under_module_control=True
        )
        self.mono_width = pc.Number(
            initial_value=500, units="wn", disable_under_module_control=True
        )
        self.mono_width.set_disabled_units(True)
        self.mono_npts = pc.Number(initial_value=51, decimals=0, disable_under_module_control=True)
        input_table = pw.InputTable()
        input_table.add("Spectrometer", None)
        input_table.add("Method", self.mono_method_combo)
        input_table.add("Center", self.mono_center)
        input_table.add("Width", self.mono_width)
        input_table.add("Number", self.mono_npts)
        self.layout.addWidget(input_table)
        self.update_mono_settings()
        # processing
        input_table = pw.InputTable()
        input_table.add("Processing", None)
        self.do_post_process = pc.Bool(initial_value=True)
        input_table.add("Process", self.do_post_process)
        # TODO: allowed values, update
        channel_names = list(yaqc_cmds.sensors.get_channels_dict().keys())
        if (
            "main_channel" not in self.state.keys()
            or self.state["main_channel"] not in channel_names
        ):
            self.state["main_channel"] = channel_names[0]
        self.main_channel = pc.Combo(
            allowed_values=channel_names,
            initial_value=self.state["main_channel"],
        )
        input_table.add("Channel", self.main_channel)
        self.layout.addWidget(input_table)

    def load(self, aqn_path):
        aqn = wt.kit.INI(aqn_path)
        # shared settings
        self.opa_combo.write(aqn.read("motortune", "opa name"))
        self.use_tune_points.write(aqn.read("motortune", "use tune points"))
        # motor settings
        opa = self.opa_guis[self.opa_combo.read_index()]
        for motor, motor_name in zip(opa.motors, aqn.read("motortune", "motor names")):
            motor.method.write(aqn.read(motor_name, "method"))
            motor.center.write(aqn.read(motor_name, "center"))
            motor.width.write(aqn.read(motor_name, "width"))
            motor.npts.write(aqn.read(motor_name, "number"))
        # mono settings
        self.mono_method_combo.write(aqn.read("spectrometer", "method"))
        self.mono_center.write(aqn.read("spectrometer", "center"))
        self.mono_width.write(aqn.read("spectrometer", "width"))
        self.mono_npts.write(aqn.read("spectrometer", "number"))
        # processing
        self.do_post_process.write(aqn.read("processing", "do post process"))
        self.main_channel.write(aqn.read("processing", "channel"))
        # allow sensors to read from aqn
        # self.device_widget.load(aqn_path)

    def on_device_settings_updated(self):
        self.main_channel.set_allowed_values(sensors.control.channel_names)

    def on_opa_combo_updated(self):
        self.show_opa_gui(self.opa_combo.read_index())

    def save(self, aqn_path):
        aqn = wt.kit.INI(aqn_path)
        opa = self.opa_guis[self.opa_combo.read_index()]
        scanned_motor_names = []
        for motor in opa.motors:
            if motor.method == "Scan":
                scanned_motor_names.append(motor.name)
        scanned_motor_names = str(scanned_motor_names).replace("'", "")
        aqn.write(
            "info",
            "description",
            "MOTORTUNE: {} {}".format(self.opa_combo.read(), scanned_motor_names),
        )
        # shared settings
        aqn.add_section("motortune")
        aqn.write("motortune", "opa name", self.opa_combo.read())
        aqn.write(
            "motortune",
            "motor names",
            [motor.name for motor in self.opa_guis[self.opa_combo.read_index()].motors],
        )
        aqn.write("motortune", "use tune points", self.use_tune_points.read())
        # motor settings
        for motor in opa.motors:
            aqn.add_section(motor.name)
            aqn.write(motor.name, "method", motor.method.read())
            aqn.write(motor.name, "center", motor.center.read())
            aqn.write(motor.name, "width", motor.width.read())
            aqn.write(motor.name, "number", motor.npts.read())
        # mono settings
        aqn.add_section("spectrometer")
        aqn.write("spectrometer", "method", self.mono_method_combo.read())
        aqn.write("spectrometer", "center", self.mono_center.read())
        aqn.write("spectrometer", "center units", self.mono_center.units)
        aqn.write("spectrometer", "width", self.mono_width.read())
        aqn.write("spectrometer", "number", self.mono_npts.read())
        # processing
        aqn.add_section("processing")
        aqn.write("processing", "do post process", self.do_post_process.read())
        aqn.write("processing", "channel", self.main_channel.read())
        # allow sensors to save to aqn
        # self.device_widget.save(aqn_path)

    def show_opa_gui(self, index):
        for gui in self.opa_guis:
            gui.hide()
        self.opa_guis[index].show()

    def update_mono_settings(self):
        self.mono_center.set_disabled(True)
        self.mono_width.set_disabled(True)
        self.mono_npts.set_disabled(True)
        method = self.mono_method_combo.read()
        if method == "Set":
            self.mono_center.set_disabled(self.use_tune_points.read())
        elif method == "Scan":
            self.mono_center.set_disabled(self.use_tune_points.read())
            self.mono_width.set_disabled(False)
            self.mono_npts.set_disabled(False)
        elif method == "Static":
            self.mono_center.set_disabled(False)

    def save_state(self):
        self.state["main_channel"] = self.channel_combo.read()
        super().save_state()


def mkGUI():
    global gui
    gui = GUI(module_name)


def load():
    return True


def get_tune_points(instrument, arrangement, scanned_motors):
    min_ = arrangement.ind_min
    max_ = arrangement.ind_max
    if not scanned_motors:
        scanned_motors = arrangement.keys()
    inds = []
    for scanned in scanned_motors:
        if scanned in arrangement.keys():
            inds += [arrangement[scanned].independent]
            continue
        for name in arrangement.keys():
            if (
                name in instrument.arrangements
                and scanned in instrument(instrument[name].ind_min, name).keys()
            ):
                inds += [arrangement[scanned].independent]
    if len(inds) > 1:
        inds = np.concatenate(inds)
    else:
        inds = inds[0]

    unique = np.unique(inds)
    tol = 1e-3 * (max_ - min_)
    diff = np.append(tol * 2, np.diff(unique))
    return unique[diff > tol]
